"""
Utilities for generating documentation for Commands

:author: Doug Skrypa
"""

from __future__ import annotations

import logging
import sys
from collections import defaultdict
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Mapping, List, Dict

from .commands import Command
from .context import Context
from .core import CommandMeta, get_params, get_parent, get_metadata
from .formatting.commands import get_formatter, NameFunc
from .formatting.restructured_text import MODULE_TEMPLATE, rst_header, rst_toc_tree

if TYPE_CHECKING:
    from .typing import Bool, PathLike, CommandCls, Strings

    Commands = Dict[str, CommandCls]

__all__ = ['render_script_rst', 'render_command_rst', 'load_commands', 'RstWriter']
log = logging.getLogger(__name__)


# region Render Script / Command RST


def render_script_rst(
    path: PathLike, top_only: Bool = True, fix_name: Bool = True, fix_name_func: NameFunc = None
) -> str:
    """Load all Commands from the file with the given path, and generate a single RST string based on those Commands"""
    commands = load_commands(path, top_only)
    return _render_commands_rst(commands, fix_name, fix_name_func)


def render_command_rst(command: CommandCls, fix_name: Bool = True, fix_name_func: NameFunc = None) -> str:
    """
    :param command: The :class:`.Command` to document
    :param fix_name: Whether the file name should be re-formatted from CamelCase / snake_case to separate Title Case
      words or not (default: True)
    :param fix_name_func: The function to call if ``fix_name`` is True instead of the default one.
    :return: The help text for the given Command, formatted using RST
    """
    with Context([], command, allow_argv_prog=False):
        return get_formatter(command).format_rst(fix_name, fix_name_func, no_sys_argv=True)


def _render_commands_rst(commands: Commands, fix_name: Bool = True, fix_name_func: NameFunc = None) -> str:
    # This could be better, but it's relatively unlikely to have multiple top level commands in a script...
    # For the same reason that main() does not try to pick one, this will just combine all of them.
    parts = []
    for i, (_name, command) in enumerate(sorted(commands.items())):
        if i:
            parts.append('\n--------\n')

        parts.append(render_command_rst(command, fix_name, fix_name_func))

    if len(parts) == 1:
        return parts[0]
    return '\n'.join(parts)


# endregion


# region Import and Load Commands


def load_commands(path: PathLike, top_only: Bool = False) -> Commands:
    """
    Load all of the commands from the file with the given path and return them as a dict of ``{name: Command}``.

    If an :class:`python:OSError` or a subclass thereof is encountered while attempting to load the file (due to the
    path not existing, or a permission error, etc), it will be allowed to propagate.  An :class:`python:ImportError`
    may be raised by :func:`import_module` if the specified path cannot be imported.

    :param path: The path to a file containing one or more :class:`.Command` classes
    :param top_only: If True, then only top-level commands are returned (default: all)
    :return: Dict containing the Commands loaded from the given file
    """
    with Context(allow_argv_prog=False):
        module = import_module(path)
    commands = {key: val for key, val in module.__dict__.items() if not key.startswith('__') and _is_command(val)}
    if top_only:
        commands = top_level_commands(commands)

    doc_str = module.__doc__
    if doc_str:
        for command in commands.values():
            get_metadata(command).pkg_doc_str = doc_str

    return commands


def top_level_commands(commands: Commands) -> Commands:
    """Filter the given commands to only the ones that do not have a parent present in the provided dict of commands"""
    if len(commands) <= 1:
        return commands

    indirect_parents = defaultdict(set)
    for name, command in commands.items():
        sub_command = get_params(command).sub_command
        if sub_command:
            for choice in sub_command.choices.values():
                indirect_parents[choice.target].add(command)

    all_commands = set(commands.values())

    filtered = {}
    for name, command in commands.items():
        parent = get_parent(command, False)
        if not ((parent and parent in all_commands) or indirect_parents[command]):
            filtered[name] = command

    return filtered


def import_module(path: PathLike):
    """Import the module / package from the given path"""
    path = Path(path)
    name = path.stem
    if path.is_dir():
        path /= '__init__.py'
    spec = spec_from_file_location(name, path)
    try:
        module = module_from_spec(spec)
    except AttributeError as e:
        path_str = path.as_posix()
        raise ImportError(f'Invalid path={path_str!r} - are you sure it is a Python module?', path=path_str) from e
    sys.modules[spec.name] = module  # This is required for the program metadata introspection
    try:
        spec.loader.exec_module(module)
    except Exception:
        del sys.modules[spec.name]
        raise
    return module


def _is_command(obj) -> bool:
    return isinstance(obj, CommandMeta) and obj is not Command


# endregion


class RstWriter:
    """
    A helper class for generating RST documentation for a Python package and/or scripts containing Commands.

    :param output_dir: Directory in which RST files should be written.
    :param dry_run: If True, log the actions that would be taken instead of taking them.
    :param encoding: The text encoding to use for output.
    :param newline: The newline character to use for output.
    :param ext: The file extension / suffix (including the leading ``.``) to use for output.
    :param module_template: The format string to use when generating RST for Python modules.
    :param skip_modules: A collection of module names (using ``package.module`` notation) that should be skipped
      when documenting a Python package via :meth:`.document_package`.
    """

    def __init__(
        self,
        output_dir: PathLike,
        *,
        dry_run: Bool = False,
        encoding: str = 'utf-8',
        newline: str = '\n',
        ext: str = '.rst',
        module_template: str = MODULE_TEMPLATE,
        skip_modules: Strings = None,
    ):
        self.output_dir = Path(output_dir)
        self.dry_run = dry_run
        self.encoding = encoding
        self.newline = newline
        self.ext = ext
        self.module_template = module_template
        self.skip_modules = set(skip_modules) if skip_modules else set()

    def document_script(
        self,
        path: Path,
        subdir: str = None,
        name: str = None,
        replacements: Mapping[str, str] = None,
        top_only: Bool = True,
        **kwargs,
    ) -> str:
        """
        Generate an RST file to document a Python script containing one or more Command classes.

        :param path: Path for a file containing one or more Command classes.
        :param subdir: If specified, write RST output for this script in this subdirectory, relative to the specified
          :paramref:`output_dir<RstWriter.output_dir>`.
        :param name: Replacement name to use as the stem of the RST file name and as the title of the page.  To replace
          the RST file name, but preserve default behavior for the page title, use ``fix_name=False`` with this param.
          The default page title is based on the name of the file that contains the Command, but can be overridden by
          providing a :ref:`configuration:Command Metadata:doc_name` value when defining the Command.
        :param replacements: A mapping of simple string replacements to apply to the generated RST content before
          saving it.  For each key=value pair, ``rst_str = rst_str.replace(key, value)`` will be performed.
        :param top_only: If True (the default), then only top-level commands in the given file will be documented,
          otherwise all commands will be documented.  When True, subcommands of the discovered top-level commands will
          still be documented.
        :param kwargs: Additional keyword arguments to pass to :func:`render_script_rst`
        :return: The stem of the file name that was used when saving the RST content for the given script.
        """
        if name:
            kwargs['fix_name_func'] = lambda n: name
            rst_name = Path(name).stem
        else:
            rst_name = path.stem

        rst_str = render_script_rst(path, top_only=top_only, **kwargs)
        if replacements:
            for key, val in replacements.items():
                rst_str = rst_str.replace(key, val)

        self.write_rst(rst_name, rst_str, subdir)
        return rst_name

    def document_scripts(
        self,
        paths: Iterable[Path],
        subdir: str = None,
        top_only: Bool = True,
        *,
        index_name: str = None,
        index_header: str = None,
        index_subdir: str = None,
        caption: str = None,
        **kwargs,
    ):
        names = [self.document_script(path, subdir, top_only=top_only, **kwargs) for path in paths]
        if index_name or index_header or index_subdir:
            name = index_name or subdir
            self.write_index(name, index_header or name.title(), names, subdir, caption, index_subdir)

    def document_module(self, module: str, subdir: str = None):
        """
        Generate an RST file to document a Python module.

        :param module: The name of the module that should be documented, using ``package.module`` notation.
        :param subdir: If specified, write RST output for the specified module in this subdirectory, relative to the
          specified :paramref:`output_dir<RstWriter.output_dir>`.
        """
        name = module.split('.')[-1].title()
        rendered = self.module_template.format(header=rst_header(f'{name} Module', 2), module=module)
        self.write_rst(module, rendered, subdir)

    def document_package(
        self,
        pkg_name: str,
        pkg_path: Path,
        subdir: str = None,
        *,
        name: str = None,
        header: str = None,
        index: Bool = True,
        empty: Bool = False,
        caption: str = None,
    ) -> List[str]:
        """
        :param pkg_name: The name of the package to document
        :param pkg_path: The path to the package
        :param subdir: The output subdirectory for package contents
        :param name: The name to use for the index file
        :param header: Header text to use in the index (default is based on the package name)
        :param index: Whether the index file should be created
        :param empty: Whether an index file should be created if the package had no modules to document
        :param caption: A caption to use for the index
        :return: List of the names from the contents of the package
        """
        if name:
            index_subdir = content_subdir = f'{subdir}/{name}' if subdir else name
        else:
            index_subdir = None
            content_subdir = subdir

        contents = self._generate_code_rsts(pkg_name, pkg_path, content_subdir)
        if (not contents and not empty) or not index:
            return contents

        if not header:
            header = '{} Package'.format(pkg_name.split('.')[-1].title())

        self.write_index(name or pkg_name, header, contents, index_subdir, caption=caption, subdir=subdir)
        return contents

    def _generate_code_rsts(self, pkg_name: str, pkg_path: Path, subdir: str = None) -> List[str]:
        contents = []
        for path in pkg_path.iterdir():
            if path.is_dir():
                sub_pkg_name = f'{pkg_name}.{path.name}'
                pkg_modules = self.document_package(sub_pkg_name, path, subdir)
                if pkg_modules:
                    contents.append(sub_pkg_name)
            elif path.is_file() and path.suffix == '.py' and not path.name.startswith('__'):
                name = f'{pkg_name}.{path.stem}'
                if name in self.skip_modules:
                    continue
                contents.append(name)
                self.document_module(name, subdir)

        return contents

    def write_index(
        self,
        name: str,
        header: str,
        contents: Strings,
        content_subdir: str = None,
        caption: str = None,
        subdir: str = None,
    ):
        content_fmt = '    {}' if content_subdir is None else f'    {content_subdir}/{{}}'
        rendered = rst_toc_tree(header, content_fmt, contents, caption=caption)
        self.write_rst(name, rendered, subdir)

    def write_rst(self, name: str, content: str, subdir: str = None):
        target_dir = self.output_dir.joinpath(subdir) if subdir else self.output_dir
        if not self.dry_run and not target_dir.exists():
            target_dir.mkdir(parents=True)

        prefix = '[DRY RUN] Would write' if self.dry_run else 'Writing'
        path = target_dir.joinpath(name + self.ext)
        log.debug(f'{prefix} {path.as_posix()}')
        if not self.dry_run:
            # Path.write_text on 3.7 does not support `newline`
            with path.open('w', encoding=self.encoding, newline=self.newline) as f:
                f.write(content)



"""
Custom date/time input handlers for Parameters.

.. warning::

    Uses :func:`python:locale.setlocale` if alternate locales are specified, which may cause problems on some systems.
    Using alternate locales in this manner should not be used in a multi-threaded application, as it will lead to
    unexpected output from other parts of the program.

    If you need to handle multiple locales and this is a problem for your application, then you should leave the
    ``locale`` parameters empty / ``None`` and use a proper i18n library like `babel <https://babel.pocoo.org/>`__
    for localization.

:author: Doug Skrypa
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from calendar import day_name, day_abbr, month_name, month_abbr
from datetime import datetime, date, time, timedelta
from enum import Enum
from locale import LC_ALL, setlocale
from threading import RLock
from typing import Union, Iterator, Collection, Sequence, Optional, TypeVar, Type, overload
from typing import Tuple, Dict

try:
    from typing import Literal
except ImportError:
    from ..compat import Literal

from ..typing import T, Bool, Locale, TimeBound
from ..utils import MissingMixin
from .base import InputType
from .exceptions import InputValidationError, InvalidChoiceError

__all__ = ['DTFormatMode', 'Day', 'Month', 'TimeDelta', 'DateTime', 'Date', 'Time']

DT = TypeVar('DT')
TimeUnit = Literal['days', 'seconds', 'microseconds', 'milliseconds', 'minutes', 'hours', 'weeks']
_TIMEDELTA_UNITS = {'days', 'seconds', 'microseconds', 'milliseconds', 'minutes', 'hours', 'weeks'}
DEFAULT_DATE_FMT = '%Y-%m-%d'
DEFAULT_TIME_FMT = '%H:%M:%S'
DEFAULT_DT_FMT = '%Y-%m-%d %H:%M:%S'


class different_locale:
    """
    Context manager that allows the temporary use of an alternate locale for date/time parsing/formatting.

    Not using :class:`python:calendar.different_locale` because it results in incorrect string encoding for some
    locales (at least on Windows).
    """

    _lock = RLock()
    __slots__ = ('locale', 'original')

    def __init__(self, locale: Optional[Locale]):
        self.locale = locale

    def __enter__(self):
        self._lock.acquire()
        locale = self.locale
        if not locale:
            return
        # locale.getlocale does not support LC_ALL, but `setlocale(LC_ALL)` with no locale to set will return a str
        # containing all of the current locale settings as `key1=val1;key2=val2;...;keyN=valN`
        self.original = setlocale(LC_ALL)
        # The calendar.different_locale implementation only calls setlocale with LC_TIME, which caused LC_CTYPE
        # to remain set to `English_United States.1252` on Windows 10, which resulted in incorrectly encoded results.
        # Using f'LC_CTYPE={locale};LC_TIME={locale}' seemed cleaner than setting LC_ALL in its entirety, but it
        # resulted in `locale.Error: unsupported locale setting` on Ubuntu/WSL.
        setlocale(LC_ALL, locale)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.locale:
            setlocale(LC_ALL, self.original)
        self._lock.release()


class DTInput(InputType[T], ABC):
    __slots__ = ('locale',)
    dt_type: str
    locale: Optional[Locale]

    def __init_subclass__(cls, dt_type: str = None, **kwargs):
        """
        :param dt_type: Used in InvalidChoiceError / ValueError messages
        """
        super().__init_subclass__(**kwargs)
        cls.dt_type = dt_type

    def __init__(self, locale: Locale = None):
        self.locale = locale

    @abstractmethod
    def choice_str(self, choice_delim: str = ',', sort_choices: bool = False) -> str:
        raise NotImplementedError

    def fix_default(self, value: Union[str, T, None]) -> Optional[T]:
        if value is None or not isinstance(value, str):
            return value
        return self(value)


# region Calendar Unit Inputs


class DTFormatMode(MissingMixin, Enum):
    FULL = 'full'                   #: The full name of a given date/time unit
    ABBREVIATION = 'abbreviation'   #: The abbreviation of a given date/time unit
    NUMERIC = 'numeric'             #: The numeric representation of a given date/time unit
    NUMERIC_ISO = 'numeric_iso'     #: The ISO numeric representation of a given date/time unit


class CalendarUnitInput(DTInput[Union[str, int]], ABC):
    __slots__ = ('full', 'abbreviation', 'numeric', 'out_format', 'out_locale')
    _formats: Dict[DTFormatMode, Sequence[Union[str, int]]]
    _min_index: int = 0

    def __init_subclass__(cls, min_index: int = 0, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._min_index = min_index

    def __init__(
        self,
        *,
        full: Bool = True,
        abbreviation: Bool = True,
        numeric: Bool = False,
        locale: Locale = None,
        out_format: Union[str, DTFormatMode] = DTFormatMode.FULL,
        out_locale: Locale = None,
    ):
        """
        Input type representing a date/time unit.

        :param full: Allow the full unit name to be provided
        :param abbreviation: Allow abbreviations of unit names to be provided
        :param numeric: Allow unit values to be specified as a decimal number
        :param locale: An alternate locale to use when parsing input
        :param out_format: A :class:`DTFormatMode` or str that matches a format mode.  Defaults to full weekday name.
        :param out_locale: Alternate locale to use for output.  Defaults to the same value as ``locale``.
        """
        if not (full or abbreviation or numeric):
            raise ValueError('At least one of full, abbreviation, or numeric must be True')
        super().__init__(locale=locale)
        self.full = full
        self.abbreviation = abbreviation
        self.numeric = numeric
        self.out_format = DTFormatMode(out_format)
        self.out_locale = out_locale or self.locale
        if self.out_format not in self._formats:
            raise ValueError(f'Unsupported out_format={self.out_format} for {self.__class__.__name__} inputs')

    def _values(self) -> Iterator[Tuple[int, str]]:
        if not self.full and not self.abbreviation:
            return
        min_index = self._min_index
        with different_locale(self.locale):
            if self.full:
                yield from enumerate(self._formats[DTFormatMode.FULL][min_index:], min_index)
            if self.abbreviation:
                yield from enumerate(self._formats[DTFormatMode.ABBREVIATION][min_index:], min_index)

    def choices(self, sort: bool = False) -> Sequence[str]:
        choices = [dow for _, dow in self._values()]
        if self.numeric:
            mode = DTFormatMode.NUMERIC_ISO if getattr(self, 'iso', False) else DTFormatMode.NUMERIC
            # fmt: off
            choices.extend(map(str, self._formats[mode][self._min_index:]))
            # fmt: on
        if sort:
            choices.sort()
        return choices

    def choice_str(self, choice_delim: str = ',', sort_choices: bool = False) -> str:
        return choice_delim.join(self.choices(sort_choices))

    def format_metavar(self, choice_delim: str = ',', sort_choices: bool = False) -> str:
        return '{' + self.choice_str(choice_delim, sort_choices) + '}'

    @abstractmethod
    def parse_numeric(self, value: str) -> int:
        raise NotImplementedError

    def parse(self, value: str) -> int:
        """
        Parse the date/time unit from the given value.

        This method does not use :meth:`python:datetime.strptime` because it is not accurate for standalone numeric
        weekdays.

        :param value: The value provided as input
        :return: The numeric unit value
        """
        if self.numeric:
            try:
                return self.parse_numeric(value)
            except InputValidationError:
                raise
            except ValueError:
                pass

        choice_map = {dow.casefold(): i for i, dow in self._values()}
        try:
            return choice_map[value.casefold()]
        except KeyError:
            pass

        raise InvalidChoiceError(value, self.choices(), self.dt_type)

    def __call__(self, value: str) -> Union[str, int]:
        normalized = self.parse(value)
        if normalized < self._min_index:
            raise InvalidChoiceError(value, self.choices(), self.dt_type)

        out_mode = self.out_format
        if out_mode in (DTFormatMode.NUMERIC, DTFormatMode.NUMERIC_ISO):
            return self._formats[out_mode][normalized]

        try:
            names_or_abbreviations = self._formats[out_mode]
        except KeyError:
            pass
        else:
            with different_locale(self.out_locale):
                return names_or_abbreviations[normalized]

        raise ValueError(f'Unexpected output format={out_mode!r} for {self.dt_type}={normalized}')


class Day(CalendarUnitInput, dt_type='day of the week'):
    __slots__ = ('iso',)
    _formats = {
        DTFormatMode.FULL: day_name,
        DTFormatMode.ABBREVIATION: day_abbr,
        DTFormatMode.NUMERIC: range(7),
        DTFormatMode.NUMERIC_ISO: range(1, 8),
    }

    @overload
    def __init__(
        self,
        *,
        full: Bool = True,
        abbreviation: Bool = True,
        numeric: Bool = False,
        iso: Bool = False,
        locale: Locale = None,
        out_format: Union[str, DTFormatMode] = DTFormatMode.FULL,
        out_locale: Locale = None,
    ):
        """
        Input type representing a day of the week.

        :param full: Allow the full day name to be provided
        :param abbreviation: Allow abbreviations of day names to be provided
        :param numeric: Allow weekdays to be specified as a decimal number
        :param iso: Ignored if ``numeric`` is False.  If True, then numeric weekdays are treated as ISO 8601 weekdays,
          where 1 is Monday and 7 is Sunday.  If False, then 0 is Monday and 6 is Sunday.
        :param locale: An alternate locale to use when parsing input
        :param out_format: A :class:`DTFormatMode` or str that matches a format mode.  Defaults to full weekday name.
        :param out_locale: Alternate locale to use for output.  Defaults to the same value as ``locale``.
        """
        ...

    def __init__(self, *, iso: Bool = False, **kwargs):
        super().__init__(**kwargs)
        self.iso = iso

    def parse_numeric(self, value: str) -> int:
        try:
            dow_num = int(value)
        except (ValueError, TypeError) as e:
            raise ValueError(f'Invalid numeric weekday={value!r}') from e

        start, stop = (1, 7) if self.iso else (0, 6)
        if start <= dow_num <= stop:
            return (dow_num - 1) if self.iso else dow_num

        with different_locale(self.locale):
            raise InputValidationError(
                f'Invalid weekday={dow_num} - expected a value'
                f' between {start} ({day_name[0]}) and {stop} ({day_name[6]})'
            )


class Month(CalendarUnitInput, dt_type='month', min_index=1):
    __slots__ = ()
    _formats = {
        DTFormatMode.FULL: month_name,
        DTFormatMode.ABBREVIATION: month_abbr,
        DTFormatMode.NUMERIC: range(13),
    }

    @overload
    def __init__(
        self,
        *,
        full: Bool = True,
        abbreviation: Bool = True,
        numeric: Bool = True,
        locale: Locale = None,
        out_format: Union[str, DTFormatMode] = DTFormatMode.FULL,
        out_locale: Locale = None,
    ):
        """
        Input type representing a month.

        :param full: Allow the full month name to be provided
        :param abbreviation: Allow abbreviations of month names to be provided
        :param numeric: Allow months to be specified as a decimal number
        :param locale: An alternate locale to use when parsing input
        :param out_format: A :class:`DTFormatMode` or str that matches a format mode.  Defaults to full month name.
        :param out_locale: Alternate locale to use for output.  Defaults to the same value as ``locale``.
        """
        ...

    def __init__(self, *, numeric: Bool = True, **kwargs):
        super().__init__(numeric=numeric, **kwargs)

    def parse_numeric(self, value: str) -> int:
        try:
            month = int(value)
        except (ValueError, TypeError) as e:
            raise ValueError(f'Invalid numeric weekday={value!r}') from e

        if 1 <= month <= 12:
            return month

        with different_locale(self.locale):
            raise InputValidationError(
                f'Invalid month={month} - expected a value between 1 ({month_name[1]}) and 12 ({month_name[12]})'
            )


# endregion


class TimeDelta(InputType[timedelta]):
    __slots__ = ('unit',)

    def __init__(self, unit: TimeUnit):
        unit = unit.lower()
        if unit not in _TIMEDELTA_UNITS:
            raise TypeError(f'Invalid unit={unit!r} - expected one of: {", ".join(sorted(_TIMEDELTA_UNITS))}')
        self.unit = unit
        # TODO: min/max params like NumRange?

    def __call__(self, value: Union[str, int, float]) -> timedelta:
        if isinstance(value, str):
            try:
                value = float(value.replace(',', '').replace('_', ''))  # allow comma or _ between thousands
            except ValueError as e:
                raise InputValidationError(
                    f'Invalid numeric {self.unit}={value!r} - expected an integer or float'
                ) from e

        return timedelta(**{self.unit: value})

    def fix_default(self, value: Union[int, float, timedelta, None]) -> Optional[timedelta]:
        if value is None or isinstance(value, timedelta):
            return value
        return self(value)

    def format_metavar(self, choice_delim: str = ',', sort_choices: bool = False) -> str:
        return f'{{{self.unit}}}'


# region Date/Time Parse Inputs


class DateTimeInput(DTInput[DT], ABC):
    formats: Collection[str]
    _type: Type[DT]
    _earliest: TimeBound = None
    _latest: TimeBound = None

    def __init_subclass__(cls, type: Type[DT], **kwargs):  # noqa
        super().__init_subclass__(dt_type=type.__name__, **kwargs)
        cls._type = type

    def __init__(
        self, formats: Collection[str], locale: Locale = None, earliest: TimeBound = None, latest: TimeBound = None
    ):
        super().__init__(locale=locale)
        self.formats = formats
        self.earliest = earliest
        self.latest = latest

    @classmethod
    def _fix_type(cls, dt: datetime) -> DT:
        try:
            return getattr(dt, cls.dt_type)()
        except AttributeError:
            return dt

    @property
    def earliest(self) -> Optional[DT]:
        return self._fix_type(normalize_dt(self._earliest))

    @earliest.setter
    def earliest(self, value: TimeBound):
        self._validate_bound_combo(value, self._latest)
        self._earliest = value

    @property
    def latest(self) -> Optional[DT]:
        return self._fix_type(normalize_dt(self._latest))

    @latest.setter
    def latest(self, value: TimeBound):
        self._validate_bound_combo(self._earliest, value)
        self._latest = value

    @classmethod
    def _validate_bound_combo(cls, raw_earliest: TimeBound, raw_latest: TimeBound):
        earliest = cls._fix_type(normalize_dt(raw_earliest))
        latest = cls._fix_type(normalize_dt(raw_latest))
        if earliest is not None and latest is not None and earliest >= latest:
            raise ValueError(
                f'Invalid combination of earliest={raw_earliest!r} and latest={raw_latest!r} values'
                f' - {dt_repr(earliest, False)} >= {dt_repr(latest, False)}'
            )

    def parse_dt(self, value: str) -> datetime:
        with different_locale(self.locale):
            for fmt in self.formats:
                try:
                    return datetime.strptime(value, fmt)
                except ValueError:
                    pass

        raise InputValidationError(
            f'Expected a {self.dt_type} matching one of the following formats: {self.choice_str()}'
        )

    def parse(self, value: str) -> DT:
        return self._fix_type(self.parse_dt(value))

    def choice_str(self, choice_delim: str = ' | ', sort_choices: bool = False) -> str:
        return choice_delim.join(sorted(self.formats) if sort_choices else self.formats)

    def format_metavar(self, choice_delim: str = ' | ', sort_choices: bool = False) -> str:
        choices = '{' + self.choice_str(choice_delim, sort_choices) + '}'
        earliest, latest = self.earliest, self.latest
        prefix = f'{dt_repr(earliest, False)} <= ' if earliest is not None else ''
        suffix = f' <= {dt_repr(latest, False)}' if latest is not None else ''
        return f'[{prefix}{choices}{suffix}]' if prefix or suffix else choices

    def _validate_bounds(self, dt: DT):
        earliest, latest = self.earliest, self.latest
        check_earliest = earliest is not None
        check_latest = latest is not None
        if not (check_earliest or check_latest):
            return
        elif (check_earliest and dt < earliest) or (check_latest and dt > latest):
            if check_earliest and check_latest:
                msg = f'between {dt_repr(earliest)} and {dt_repr(latest)} (inclusive)'
            else:
                msg = f'after {dt_repr(earliest)}' if check_earliest else f'before {dt_repr(latest)}'
            raise InputValidationError(f'Invalid {self.dt_type}={dt_repr(dt)} - a {self.dt_type} {msg} is required')

    def __call__(self, value: str) -> DT:
        parsed = self.parse(value)
        self._validate_bounds(parsed)
        return parsed


class DateTime(DateTimeInput[datetime], type=datetime):
    def __init__(self, *formats: str, locale: Locale = None, earliest: TimeBound = None, latest: TimeBound = None):
        """
        Input type that accepts any number of datetime format strings for parsing input.  Parsing results in returning
        a :class:`python:datetime.datetime` object.

        :param formats: One or more :ref:`datetime format strings <python:strftime-strptime-behavior>`.  Defaults to
          :data:`DEFAULT_DT_FMT`.
        :param locale: An alternate locale to use when parsing input
        :param earliest: If specified, the parsed value must be later than or equal to this
        :param latest: If specified, the parsed value must be earlier than or equal to this
        """
        super().__init__(formats or (DEFAULT_DT_FMT,), locale=locale, earliest=earliest, latest=latest)


class Date(DateTimeInput[date], type=date):
    def __init__(self, *formats: str, locale: Locale = None, earliest: TimeBound = None, latest: TimeBound = None):
        """
        Input type that accepts any number of datetime format strings for parsing input.  Parsing results in returning
        a :class:`python:datetime.date` object.

        :param formats: One or more :ref:`datetime format strings <python:strftime-strptime-behavior>`.  Defaults to
          :data:`DEFAULT_DT_FMT`.
        :param locale: An alternate locale to use when parsing input
        :param earliest: If specified, the parsed value must be later than or equal to this
        :param latest: If specified, the parsed value must be earlier than or equal to this
        """
        super().__init__(formats or (DEFAULT_DATE_FMT,), locale=locale, earliest=earliest, latest=latest)


class Time(DateTimeInput[time], type=time):
    def __init__(self, *formats: str, locale: Locale = None, earliest: TimeBound = None, latest: TimeBound = None):
        """
        Input type that accepts any number of datetime format strings for parsing input.  Parsing results in returning
        a :class:`python:datetime.time` object.

        :param formats: One or more :ref:`datetime format strings <python:strftime-strptime-behavior>`.  Defaults to
          :data:`DEFAULT_DT_FMT`.
        :param locale: An alternate locale to use when parsing input
        :param earliest: If specified, the parsed value must be later than or equal to this
        :param latest: If specified, the parsed value must be earlier than or equal to this
        """
        super().__init__(formats or (DEFAULT_TIME_FMT,), locale=locale, earliest=earliest, latest=latest)


# endregion


def dt_repr(dt: Union[datetime, date, time], use_repr: bool = True) -> str:
    try:
        dt_str = dt.isoformat(' ')
    except (TypeError, ValueError):  # TypeError for date objects, ValueError for time objects
        dt_str = dt.isoformat()
    return repr(dt_str) if use_repr else dt_str


def normalize_dt(value: TimeBound, now: datetime = None) -> Optional[datetime]:
    if value is None or isinstance(value, datetime):
        return value
    elif isinstance(value, timedelta):
        if now is None:
            now = datetime.now()
        return now + value
    elif isinstance(value, date):
        return datetime.combine(value, time())
    elif isinstance(value, time):
        today = date.today() if now is None else now.date()
        return datetime.combine(today, value)
    raise TypeError(
        f'Unexpected datetime specifier type={value.__class__.__name__} for value={value!r}'
        ' (expected datetime, date, time, timedelta, or None)'
    )



from __future__ import annotations

import ast
import logging
import sys
from argparse import ArgumentParser
from ast import AST, Assign, Call, withitem
from functools import partial
from inspect import Signature, BoundArguments
from pathlib import Path
from typing import TYPE_CHECKING, Union, Optional, Callable, Collection, TypeVar, Generic, Type, Iterator
from typing import List, Tuple, Dict, Set

from cli_command_parser.compat import cached_property
from .argparse_utils import ArgumentParser as _ArgumentParser, SubParsersAction as _SubParsersAction
from .utils import get_name_repr, iter_module_parents, unparse

if TYPE_CHECKING:
    from cli_command_parser.typing import PathLike
    from .visitor import TrackedRefMap, TrackedRef

__all__ = ['ParserArg', 'ArgGroup', 'MutuallyExclusiveGroup', 'AstArgumentParser', 'SubParser', 'Script']
log = logging.getLogger(__name__)

InitNode = Union[Call, Assign, withitem]
OptCall = Optional[Call]
ParserCls = Type['AstArgumentParser']
ParserObj = TypeVar('ParserObj', bound='AstArgumentParser')
RepresentedCallable = TypeVar('RepresentedCallable', bound=Callable)
AC = TypeVar('AC', bound='AstCallable')
D = TypeVar('D')
_NotSet = object()


class Script:
    _parser_classes = {}
    path: Optional[Path]

    def __init__(self, src_text: str, smart_loop_handling: bool = True, path: PathLike = None):
        self.smart_loop_handling = smart_loop_handling
        self._parsers = []
        self.path = Path(path) if path else None
        self.src_text = src_text
        parse_args = (self.src_text, self.path.as_posix()) if self.path else (self.src_text,)
        self.root_node = ast.parse(*parse_args)

    def __repr__(self) -> str:
        parsers = len(self.parsers)
        location = f' @ {self.path.as_posix()}' if self.path else ''
        return f'<{self.__class__.__name__}[parsers={parsers!r}{location}]>'

    @property
    def mod_cls_to_ast_cls_map(self) -> Dict[str, Dict[str, ParserCls]]:
        return self._parser_classes

    @classmethod
    def _register_parser(cls, module: str, name: str, ast_cls: ParserCls):
        # Identify package-level exports that may have been defined for a custom ArgumentParser subclass
        modules = [module, *(parent for parent in iter_module_parents(module) if name in vars(sys.modules[parent]))]
        for module in modules:
            log.debug(f'Registering {module}.{name} -> {ast_cls}')
            cls._parser_classes.setdefault(module, {})[name] = ast_cls

    @classmethod
    def register_parser(cls, ast_cls: ParserCls):
        real_cls = ast_cls.represents
        cls._register_parser(real_cls.__module__, real_cls.__name__, ast_cls)
        return ast_cls

    def add_parser(self, ast_cls: ParserCls, node: InitNode, call: OptCall, tracked_refs: TrackedRefMap) -> ParserObj:
        parser = ast_cls(node, self, tracked_refs, call)
        self._parsers.append(parser)
        return parser

    @cached_property
    def parsers(self) -> List[ParserObj]:
        from .visitor import ScriptVisitor, TrackedRef  # noqa: F811

        track_refs = (TrackedRef('argparse.REMAINDER'), TrackedRef('argparse.SUPPRESS'))
        visitor = ScriptVisitor(self.smart_loop_handling, track_refs=track_refs)
        for module, name_cls_map in self.mod_cls_to_ast_cls_map.items():
            for name, ast_cls in name_cls_map.items():
                visitor.track_callable(module, name, partial(self.add_parser, ast_cls))

        visitor.visit(self.root_node)
        return self._parsers


# region Decorators & Descriptors


class visit_func:
    """A method that can be called by an AST visitor."""

    __slots__ = ('func',)

    def __init__(self, func):
        self.func = func

    def __set_name__(self, owner: Type[AstCallable], name: str):
        if owner._add_visit_func(name):      # This check is only to enable a low-value unit test...
            setattr(owner, name, self.func)  # There's no need to keep the descriptor - replace self with func

    def __get__(self, instance, owner):
        # This will never actually be called, but it makes PyCharm happy
        return self if instance is None else partial(self.func, instance)


class AddVisitedChild(Generic[AC]):
    """Simplifies the definition of an add_child method that can be called by an AST visitor, where possible."""

    __slots__ = ('child_cls', 'list_attr')

    def __init__(self, child_cls: Type[AC], attr: str):
        self.child_cls = child_cls
        self.list_attr = attr

    def __set_name__(self, owner: Type[ArgCollection], name: str):
        owner._add_visit_func(name)

    def __get__(self, instance: ArgCollection, owner) -> Callable[[InitNode, Call, TrackedRefMap], AC]:
        if instance is None:
            return self  # noqa
        return partial(instance._add_child, self.child_cls, getattr(instance, self.list_attr))  # noqa


# endregion


class AstCallable:
    represents: RepresentedCallable
    visit_funcs = set()
    _sig: Signature | None = None

    @classmethod
    def _add_visit_func(cls, name: str) -> bool:
        try:
            parent_visit_funcs = cls.__base__.visit_funcs  # noqa
        except AttributeError:
            pass
        else:  # Note: __init_subclass__ is called after __set_name__ is called for members
            if parent_visit_funcs is cls.visit_funcs:
                cls.visit_funcs = cls.visit_funcs.copy()
        cls.visit_funcs.add(name)
        return True

    def __init_subclass__(cls, represents: RepresentedCallable = None, **kwargs):
        super().__init_subclass__(**kwargs)
        if represents:
            cls.represents = represents
            cls._sig = None

    def __init__(self, node: InitNode, parent: AstCallable | Script, tracked_refs: TrackedRefMap, call: Call = None):
        self.init_node = node
        if not call:
            call = node.value if isinstance(node, Assign) else node  # type: Call
        self.call_node = call
        self.call_args = call.args
        self.call_kwargs = call.keywords
        self._tracked_refs = tracked_refs
        self.parent = parent

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}[{self.init_call_repr()}]>'

    def get_tracked_refs(self, module: str, name: str, default: D = _NotSet) -> Union[Set[str], D]:
        for tracked_ref, refs in self._tracked_refs.items():
            if tracked_ref.module == module and tracked_ref.name == name:
                return refs
        if default is not _NotSet:
            return default
        raise KeyError(f'No tracked ref found for {module}.{name}')

    # region Initialization Call

    @classmethod
    def _signature(cls) -> Signature:
        if not cls._sig:
            cls._sig = Signature.from_callable(cls.represents)
        return cls._sig

    @property
    def signature(self) -> Signature:
        return self._signature()

    @cached_property
    def init_func_name(self) -> str:
        """The name or alias of the function/callable that was used to initialize this object"""
        return get_name_repr(self.call_node.func)

    @cached_property
    def _init_func_bound(self) -> BoundArguments:
        args = self.call_args if isinstance(self.represents, type) else ('self', *self.call_args)
        return self.signature.bind(*args, **{kw.arg: kw.value for kw in self.call_kwargs})

    @cached_property
    def init_func_args(self) -> List[str]:
        try:
            args = self._init_func_bound.args[1:]
        except (TypeError, AttributeError):  # No represents func
            args = self.call_args
        return [unparse(arg) for arg in args]

    @cached_property
    def init_func_raw_kwargs(self) -> Dict[str, AST]:
        try:
            kwargs = self._init_func_bound.arguments
        except (TypeError, AttributeError):  # No represents func
            return {kw.arg: kw.value for kw in self.call_kwargs}
        else:
            kwargs = kwargs.copy()
            kwargs.pop('self', None)
            if isinstance(kwargs.get('args'), tuple):
                kwargs.pop('args')
            if isinstance(kwargs.get('kwargs'), dict):
                kwargs.update(kwargs.pop('kwargs'))
            return kwargs

    def _init_func_kwargs(self) -> Dict[str, str]:
        return {key: unparse(val) for key, val in self.init_func_raw_kwargs.items()}

    @cached_property
    def init_func_kwargs(self) -> Dict[str, str]:
        return self._init_func_kwargs()

    def init_call_repr(self) -> str:
        arg_str = ', '.join(self.init_func_args)
        kw_str = ', '.join(f'{k}={v}' for k, v in self.init_func_kwargs.items())
        if kw_str:
            arg_str = kw_str if not arg_str else (arg_str + ', ' + kw_str)
        return f'{self.init_func_name}({arg_str})'

    # endregion

    def pprint(self, indent: int = 0):
        print(f'{" " * indent} - {self!r}')


# region Stdlib Argparse Wrappers


class ParserArg(AstCallable, represents=ArgumentParser.add_argument):
    parent: ArgCollection


class ArgCollection(AstCallable):
    parent: ArgCollection | Script
    _children = ('args', 'groups')
    args: List[ParserArg]
    groups: List[ArgGroup]
    add_argument = AddVisitedChild(ParserArg, 'args')

    def __init_subclass__(cls, children: Collection[str] = (), **kwargs):
        super().__init_subclass__(**kwargs)
        if children:
            cls._children = (*cls._children, *children)

    def __init__(self, node: InitNode, parent: AstCallable | Script, tracked_refs: TrackedRefMap, call: Call = None):
        super().__init__(node, parent, tracked_refs, call)
        self.args = []
        self.groups = []

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: ``{self.init_call_repr()}``>'

    def _add_child(self, cls: Type[AC], container: List[AC], node: InitNode, call: Call, refs: TrackedRefMap) -> AC:
        child = cls(node, self, refs, call)
        container.append(child)
        return child

    @visit_func
    def add_mutually_exclusive_group(self, node: InitNode, call: Call, tracked_refs: TrackedRefMap):
        return self._add_child(MutuallyExclusiveGroup, self.groups, node, call, tracked_refs)

    @visit_func
    def add_argument_group(self, node: InitNode, call: Call, tracked_refs: TrackedRefMap):
        return self._add_child(ArgGroup, self.groups, node, call, tracked_refs)

    def grouped_children(self) -> Iterator[Tuple[Type[AC], List[AC]]]:
        yield ParserArg, self.args
        yield ArgGroup, self.groups

    # region Output Methods

    def pprint(self, indent: int = 0):
        print(f'{" " * indent} + {self!r}:')
        indent += 3
        for attr in self._children:
            values = getattr(self, attr)
            if values:
                for value in values:
                    value.pprint(indent)

    # endregion


class ArgGroup(ArgCollection, represents=_ArgumentParser.add_argument_group):
    pass


class MutuallyExclusiveGroup(ArgGroup, represents=_ArgumentParser.add_mutually_exclusive_group):
    pass


class SubparsersAction(AstCallable, represents=_ArgumentParser.add_subparsers):
    parent: ParserObj

    @visit_func
    def add_parser(self, node: InitNode, call: Call, tracked_refs: TrackedRefMap):
        sub_parser = self.parent._add_subparser(node, call, tracked_refs)
        sub_parser.sp_parent = self
        return sub_parser


@Script.register_parser
class AstArgumentParser(ArgCollection, represents=ArgumentParser, children=('sub_parsers',)):
    sub_parsers: List[SubParser]
    add_subparsers = AddVisitedChild(SubparsersAction, '_subparsers_actions')

    def __init__(self, node: InitNode, parent: AstCallable | Script, tracked_refs: TrackedRefMap, call: Call = None):
        super().__init__(node, parent, tracked_refs, call)
        self._subparsers_actions = []
        # Note: sub_parsers aren't included in grouped_children since they need different handling during conversion
        self.sub_parsers = []

    def __repr__(self) -> str:
        sub_parsers = len(self.sub_parsers)
        return f'<{self.__class__.__name__}[sub_parsers={sub_parsers!r}]: ``{self.init_call_repr()}``>'

    def _add_subparser(self, node: InitNode, call: Call, tracked_refs: TrackedRefMap, sub_parser_cls: ParserCls = None):
        # Using default of None since the class hasn't been defined at the time it would need to be set as default
        return self._add_child(sub_parser_cls or SubParser, self.sub_parsers, node, call, tracked_refs)


class SubParser(AstArgumentParser, represents=_SubParsersAction.add_parser):
    sp_parent: SubparsersAction

    @cached_property
    def init_func_kwargs(self) -> Dict[str, str]:
        kwargs = self.sp_parent.init_func_kwargs.copy()
        kwargs.update(self._init_func_kwargs())
        return kwargs


# endregion



from __future__ import annotations

from ast import AST, expr, Call, Attribute, Name, Dict, List, Set, Tuple

try:
    from ast import unparse
except ImportError:  # added in 3.9
    try:
        from astunparse import unparse as _unparse
    except ImportError as e:
        raise RuntimeError(
            'Missing required dependency: astunparse (only required in Python 3.8 and below'
            ' - upgrade to 3.9 or above to avoid this dependency)'
        )
    else:

        def unparse(node):
            return ''.join(_unparse(node).splitlines())


from typing import Union, Iterator, List as _List

__all__ = ['get_name_repr', 'iter_module_parents', 'collection_contents']


def get_name_repr(node: Union[AST, expr]) -> str:
    if isinstance(node, Call):
        node = node.func

    if isinstance(node, Name):
        return node.id
    elif isinstance(node, Attribute):
        return f'{get_name_repr(node.value)}.{node.attr}'  # noqa
    elif isinstance(node, AST):
        return unparse(node)
    else:
        raise TypeError(f'Only AST nodes are supported - found {node.__class__.__name__}')


def iter_module_parents(module: str) -> Iterator[str]:
    while True:
        parent = module.rsplit('.', 1)[0]
        if parent == module:
            break
        yield parent
        module = parent


def collection_contents(node: AST) -> _List[str]:
    if isinstance(node, Dict):
        return [unparse(key) for key in node.keys]  # noqa
    elif isinstance(node, (List, Set, Tuple)):
        return [unparse(ele) for ele in node.elts]  # noqa
    else:
        raise TypeError(f'Unexpected AST node type={node.__class__.__name__}')



"""
Custom numeric input handlers for Parameters

:author: Doug Skrypa
"""
# pylint: disable=W0622

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union, Optional

from ..typing import Bool, NT, Number, NumType, RngType
from .base import InputType

__all__ = ['Range', 'NumRange']

_range = range


class NumericInput(InputType[NT], ABC):
    __slots__ = ()
    type: NumType

    def is_valid_type(self, value: str) -> bool:
        try:
            self.type(value)
        except (ValueError, TypeError):
            return False
        return True

    @abstractmethod
    def _range_str(self, var: str = 'N') -> str:
        raise NotImplementedError

    def format_metavar(self, choice_delim: str = ',', sort_choices: bool = False) -> str:
        return f'{{{self._range_str()}}}'

    def fix_default(self, value: Union[str, NT, None]) -> Optional[NT]:
        if value is None or not isinstance(value, str):
            return value
        return self(value)


class Range(NumericInput[NT]):
    """
    A range of integers that uses the builtin :class:`python:range`.  If a range object is passed to a
    :class:`.Parameter` as the ``type=`` value, it will automatically be wrapped by this class.

    :param range: A :class:`python:range` object
    :param snap: If True and a provided value is outside the allowed range, snap to the nearest bound.  The min or max
      of the provided range (not necessarily the start/stop values) will be used, depending on which one the provided
      value is closer to.
    """

    type: NumType = int
    range: Optional[_range]
    snap: bool

    def __init__(self, range: RngType, snap: Bool = False, type: NumType = None):  # noqa
        self.snap = snap
        if isinstance(range, int):
            self.range = _range(range)
        elif not isinstance(range, _range):
            self.range = _range(*range)
        else:
            self.range = range
        if type is not None:
            self.type = type

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}({self.range!r}, snap={self.snap!r}, type={self.type!r})>'

    def _range_str(self, var: str = 'N') -> str:
        rng_min, rng_max = min(self.range), max(self.range)
        step = abs(self.range.step)
        base = f'{rng_min} <= {var} <= {rng_max}'
        return base if step == 1 else f'{base}, step={step}'

    def __call__(self, value: str) -> NT:
        value = self.type(value)
        if value in self.range:
            return value
        elif self.snap:
            rng_min = min(self.range)
            if value < rng_min:
                return rng_min
            return max(self.range)

        raise ValueError(f'expected a value in the range {self._range_str()}')


class NumRange(NumericInput[NT]):
    """
    A range of integers or floats, optionally only bounded on one side.

    By default, the min and max behave like the builtin :class:`python:range` - the min is inclusive, and the max
    is exclusive.

    :param type: The type for values, or any callable that returns an int/float
    :param snap: If True and a provided value is outside the allowed range, snap to the nearest bound.  Respects
      inclusivity/exclusivity of the bound.  Not supported for floats since there is not an obviously correct
      behavior for handling them in this context.
    :param min: The minimum allowed value, or None to have no lower bound
    :param max: The maximum allowed value, or None to have no upper bound
    :param include_min: Whether the minimum is inclusive (default: True)
    :param include_max: Whether the maximum is inclusive (default: False)
    """

    __slots__ = ('type', 'snap', 'min', 'max', 'include_min', 'include_max')
    snap: bool
    min: Number
    max: Number
    include_min: bool
    include_max: bool

    def __init__(
        self,
        type: NumType = None,  # noqa
        snap: Bool = False,
        *,
        min: Number = None,  # noqa
        max: Number = None,  # noqa
        include_min: Bool = True,
        include_max: Bool = False,
    ):
        if min is None and max is None:
            raise ValueError('NumRange inputs must be initialized with at least one of min and/or max values')
        elif min is not None and max is not None and min >= max:
            raise ValueError(f'Invalid min={min} >= max={max} - min must be less than max')

        if type is None:
            self.type = float if isinstance(min, float) or isinstance(max, float) else int
        else:
            self.type = type

        if snap:
            if self.type is float:
                raise TypeError('Unable to snap to extrema with type=float')
            real_min = min if include_min else min + 1
            real_max = max if include_max else max - 1
            if real_min >= real_max:
                raise ValueError(
                    f'Invalid min={min} >= max={max} with snap=True, include_min={include_min},'
                    f' include_max={include_max} - snap would produce invalid values'
                )

        self.snap = snap
        self.min = self.type(min) if min is not None else min   # for floats especially, such as a range like 0~1, this
        self.max = self.type(max) if max is not None else max   # helps to highlight the type in reprs
        self.include_min = include_min
        self.include_max = include_max

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}({self.type!r}, snap={self.snap!r})[{self._range_str()}]>'

    def _range_str(self, var: str = 'N') -> str:
        if self.min is not None:
            min_str = f'{self.min} {"<=" if self.include_min else "<"} '
        else:
            min_str = ''

        if self.max is not None:
            max_str = f' {"<=" if self.include_max else "<"} {self.max}'
        else:
            max_str = ''

        return f'{min_str}{var}{max_str}'

    def handle_invalid(self, bound: Number, inclusive: bool, snap_dir: int) -> Number:
        """
        Handle calculating / returning a snap value or raise an exception if snapping to the bound is not allowed.

        May be overridden in a subclass to support snapping for float values or other behavior.

        :param bound: The bound (min or max) that was violated
        :param inclusive: True if the bound is inclusive, False if it is exclusive
        :param snap_dir: The direction to adjust the bound if it is exclusive as ``+1`` or ``-1``
        :return: The snap value if :attr:`.snap` is True, otherwise a :class:`python:ValueError` is raised
        """
        if self.snap:
            return bound if inclusive else (bound + snap_dir)
        raise ValueError(f'expected a value in the range {self._range_str()}')

    def __call__(self, value: str) -> NT:
        value = self.type(value)
        if self.min is not None:
            # Bad if < when inclusive, bad if <= when exclusive
            if (value < self.min) if self.include_min else (value <= self.min):
                return self.handle_invalid(self.min, self.include_min, 1)
        if self.max is not None:
            if (value > self.max) if self.include_max else (value >= self.max):
                return self.handle_invalid(self.max, self.include_max, -1)
        return value



"""
Utils for input types

:author: Doug Skrypa
"""

from __future__ import annotations

import sys
import warnings
from contextlib import contextmanager
from pathlib import Path
from stat import S_IFMT, S_IFDIR, S_IFCHR, S_IFBLK, S_IFREG, S_IFIFO, S_IFLNK, S_IFSOCK
from typing import TYPE_CHECKING, Union, Any, TextIO, BinaryIO, ContextManager
from weakref import finalize

from ..utils import FixedFlag
from .exceptions import InputValidationError

if TYPE_CHECKING:
    from ..typing import Bool, FP, Converter

__all__ = ['InputParam', 'StatMode', 'FileWrapper', 'fix_windows_path']


class InputParam:
    __slots__ = ('default', 'name')

    def __init__(self, default: Any):
        self.default = default

    def __set_name__(self, owner, name: str):
        self.name = name

    def __get__(self, instance, owner) -> Any:
        if instance is None:
            return self
        try:
            return instance.__dict__[self.name]
        except KeyError:
            return self.default

    def __set__(self, instance, value: Any):
        if value != self.default:
            instance.__dict__[self.name] = value


class StatMode(FixedFlag):
    def __new__(cls, mode, friendly_name: str = None):
        # Defined __new__ to avoid juggling dicts for the stat mode values and names
        obj = object.__new__(cls)
        if friendly_name:
            obj.mode = mode
            obj.friendly_name = friendly_name
            if mode is None:  # ANY
                obj._value_ = sum(m._value_ for m in cls.__members__.values())
            else:
                obj._value_ = 2 ** len(cls.__members__)
        return obj

    DIR = S_IFDIR, 'directory'
    FILE = S_IFREG, 'regular file'
    CHARACTER = S_IFCHR, 'character special device file'
    BLOCK = S_IFBLK, 'block special device file'
    FIFO = S_IFIFO, 'FIFO (named pipe)'
    LINK = S_IFLNK, 'symbolic link'
    SOCKET = S_IFSOCK, 'socket'
    ANY = None, 'any'

    def matches(self, mode: int) -> bool:
        mode = S_IFMT(mode)
        return any(mode == part.mode for part in self._decompose())

    def __str__(self) -> str:
        try:
            name = self.friendly_name
        except AttributeError:  # Combined flags
            name = None
        if name:
            return name
        names = [part.friendly_name for part in self._decompose()]
        if len(names) == 2:
            return '{} or {}'.format(*names)
        names[-1] = f'or {names[-1]}'
        return ', '.join(names)


class FileWrapper:
    def __init__(
        self,
        path: Path,
        mode: str = 'r',
        encoding: str = None,
        errors: str = None,
        converter: Converter = None,
        pass_file: Bool = False,
        parents: Bool = False,
    ):
        self.path = path
        self.mode = mode
        self.binary = 'b' in mode
        self.encoding = encoding
        self.errors = errors
        self.converter = converter
        self.pass_file = pass_file
        self.parents = parents
        self._fp: Union[TextIO, BinaryIO, None] = None
        self._finalizer = None

    def __eq__(self, other: FileWrapper) -> bool:
        attrs = ('path', 'mode', 'binary', 'encoding', 'errors', 'converter', 'pass_file')
        try:
            return all(getattr(self, a) == getattr(other, a) for a in attrs)
        except AttributeError:
            return NotImplemented

    def read(self) -> Any:
        with self._file() as f:
            if self.converter is not None:
                return self.converter(f if self.pass_file else f.read())
            else:
                return f.read()

    def write(self, data: Any):
        with self._file() as f:
            if self.converter is not None:
                if self.pass_file:
                    self.converter(data, f)
                else:
                    f.write(self.converter(data))
            else:
                f.write(data)

    def _open(self) -> FP:
        if self.path == Path('-'):
            stream = sys.stdin if 'r' in self.mode else sys.stdout
            return stream.buffer if self.binary else stream

        if self.parents and allows_write(self.mode):
            self.path.parent.mkdir(parents=True, exist_ok=True)

        try:
            self._fp = fp = self.path.open(self.mode, encoding=self.encoding, errors=self.errors)
        except OSError as e:
            raise InputValidationError(f'Unable to open {self.path} - {e}') from e
        else:
            self._finalizer = finalize(self, self._cleanup, fp, f'Implicitly cleaning up {self.path}')
            return fp

    @classmethod
    def _cleanup(cls, fp: FP, warn_msg: str):
        fp.close()
        warnings.warn(warn_msg, ResourceWarning)

    def _close(self):
        try:
            self._fp.close()
        except AttributeError:
            pass
        finally:
            self._fp = None

    def close(self):
        try:
            do_close = self._finalizer.detach()
        except AttributeError:
            do_close = False
        if do_close:
            self._close()

    @contextmanager
    def _file(self) -> ContextManager[FP]:
        try:
            yield self._open()
        finally:
            self.close()

    def __enter__(self) -> Union[FP, FileWrapper]:
        if self.converter is not None:
            return self
        return self._open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def allows_write(mode: str, strict: bool = False) -> bool:
    chars = 'wxa' if strict else 'wxa+'
    return any(c in mode for c in chars)


def fix_windows_path(path: Path) -> Path:
    """
    Attempts to resolve issues related to inconsistencies between the way the version of Bash that is distributed with
    Git handles paths in some situations and the way that Python handles paths.

    The use case that this function currently handles is when the given Path does not exist, and it was auto-completed
    by Git Bash to begin with ``/{drive}/...`` instead of ``{drive}:/...``.
    """
    if path.exists() or not path.as_posix().startswith('/'):
        return path

    try:
        _, drive_letter, *parts = path.parts
    except ValueError:
        return path

    if len(drive_letter) != 1:
        return path

    drive = drive_letter.upper() + ':/'
    alt_path = Path(drive, *parts)
    if alt_path.exists() or (Path(drive).exists() and not Path(f'/{drive_letter}/').exists()):
        return alt_path
    else:
        return path



from __future__ import annotations

import logging
from ast import NodeVisitor, AST, Assign, Call, For, Attribute, Name, Import, ImportFrom, expr
from collections import ChainMap, defaultdict
from functools import partial, wraps
from typing import Iterator, Union, Callable, Collection, Tuple, List, Dict, Set

from .argparse_ast import AstArgumentParser
from .utils import get_name_repr

__all__ = ['ScriptVisitor', 'TrackedRef']
log = logging.getLogger(__name__)

TrackedRefMap = Dict['TrackedRef', Set[str]]
_NoCall = object()


def scoped(func):
    @wraps(func)
    def _scoped_method(self: ScriptVisitor, *args, **kwargs):
        self.scopes = self.scopes.new_child()
        try:
            func(self, *args, **kwargs)
        finally:
            self.scopes = self.scopes.parents

    return _scoped_method


class ScopedVisit:
    __slots__ = ()

    def __get__(self, instance: ScriptVisitor, owner):
        return self if instance is None else partial(scoped(owner.generic_visit), instance)


class ScriptVisitor(NodeVisitor):
    visit_FunctionDef = visit_AsyncFunctionDef = ScopedVisit()
    visit_Lambda = ScopedVisit()
    visit_ClassDef = ScopedVisit()
    visit_While = ScopedVisit()

    def __init__(self, smart_loop_handling: bool = True, track_refs: Collection[TrackedRef] = ()):
        self.smart_loop_handling = smart_loop_handling
        self.scopes = ChainMap()
        self._tracked_refs = set()
        self._mod_name_tracked_map = defaultdict(dict)
        for ref in track_refs:
            self.track_refs_to(ref)

    def track_callable(self, module: str, name: str, cb: Callable):
        self._mod_name_tracked_map[module][name] = cb

    def track_refs_to(self, ref: TrackedRef):
        self._tracked_refs.add(ref)
        self._mod_name_tracked_map[ref.module][ref.name] = ref

    def get_tracked_refs(self) -> TrackedRefMap:
        tracked_refs = defaultdict(set)
        for key, val in self.scopes.items():
            if val in self._tracked_refs:
                tracked_refs[val].add(key)
        tracked_refs.default_factory = None
        return tracked_refs

    # region Imports

    def visit_Import(self, node: Import):
        for module_name, as_name in imp_names(node):
            name_tracked_map = self._mod_name_tracked_map.get(module_name)
            if name_tracked_map:
                log.debug(f'Found module import: {module_name} as {as_name}')
                for name, tracked in name_tracked_map.items():
                    self.scopes[f'{as_name}.{name}'] = tracked

    def visit_ImportFrom(self, node: ImportFrom):
        name_tracked_map = self._mod_name_tracked_map.get(node.module)
        if name_tracked_map:
            for name, as_name in imp_names(node):
                tracked = name_tracked_map.get(name)
                if tracked:
                    log.debug(f'Found tracked import: {node.module}.{name} as {as_name}')
                    self.scopes[as_name] = tracked

    # endregion

    # region Scope Changes

    @scoped
    def visit_For(self, node: For):
        if isinstance(node.target, Name):
            try:
                ele_names = [get_name_repr(ele) for ele in node.iter.elts]  # noqa
            except (AttributeError, TypeError):
                ele_names = ()

            if ele_names and self.smart_loop_handling:
                self._visit_for_smart(node, node.target.id, ele_names)
            else:
                self._visit_for_elements(node, node.target.id, ele_names)
        else:
            self.generic_visit(node)

    visit_AsyncFor = visit_For

    def _visit_for_smart(self, node: For, loop_var: str, ele_names: List[str]):
        log.debug(f'Attempting smart for loop visit for loop_var={loop_var!r} in ele_names={ele_names!r}')
        refs = [ref for ref in (self.scopes.get(name) for name in ele_names) if ref]
        # log.debug(f'  > Found {len(refs)=}, {len(ele_names)=}')

        if len(refs) == len(ele_names) and all(isinstance(ref, AstArgumentParser) for ref in refs):
            parents = set(ref.parent for ref in refs)
            log.debug(f'  > Found parents={len(parents)}')
            if len(parents) == 1:
                parent = next(iter(parents))
                if parent and set(getattr(parent, 'sub_parsers', ())) == set(refs):
                    self.scopes[loop_var] = parent
                    self.generic_visit(node)
                    return

        log.debug('Falling back to generic loop handling')
        self._visit_for_elements(node, loop_var, ele_names)

    def _visit_for_elements(self, node: For, loop_var: str, ele_names: List[str]):
        visited_any = False
        for name in ele_names:
            ref = self.scopes.get(name)
            if ref:
                visited_any = True
                self.scopes[loop_var] = ref
                self.generic_visit(node)

        if not visited_any:
            self.generic_visit(node)

    # endregion

    def resolve_ref(self, name: Union[str, AST, Attribute, Name, expr]):
        if not isinstance(name, str):
            name = get_name_repr(name)
        try:
            return self.scopes[name]
        except KeyError:
            pass
        try:
            obj_name, attr = name.rsplit('.', 1)
        except ValueError:
            return None
        try:
            obj = self.scopes[obj_name]
        except KeyError:
            return None
        try:
            can_call = attr in obj.visit_funcs
        except (AttributeError, TypeError):
            return None
        return getattr(obj, attr) if can_call else None

    def visit_withitem(self, item):
        context_expr = item.context_expr
        func = self.resolve_ref(context_expr)
        if func:
            call = context_expr if isinstance(context_expr, Call) else None
            result = func(item, call, self.get_tracked_refs())
            as_name = item.optional_vars
            if as_name:
                self.scopes[get_name_repr(as_name)] = result

    def visit_Assign(self, node: Assign):
        value = node.value
        if isinstance(value, (Attribute, Name)):  # Assigning an alias to a variable
            ref = self.resolve_ref(value)
            if ref:
                for target in node.targets:
                    self.scopes[get_name_repr(target)] = ref
        elif isinstance(value, Call):
            result = self.visit_Call(value)
            if result is not _NoCall:
                for target in node.targets:
                    self.scopes[get_name_repr(target)] = result  # noq

    def visit_Call(self, node: Call):
        func = self.resolve_ref(node.func)
        if func:
            return func(node, node, self.get_tracked_refs())
        return _NoCall


class TrackedRef:
    __slots__ = ('module', 'name')

    def __init__(self, name: str):
        self.module, self.name = name.rsplit('.', 1)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.module}.{self.name}>'

    def __hash__(self) -> int:
        return hash(self.__class__) ^ hash(self.module) ^ hash(self.name)

    def __eq__(self, other: TrackedRef) -> bool:
        return self.name == other.name and self.module == other.module


def imp_names(imp: Import | ImportFrom) -> Iterator[Tuple[str, str]]:
    for alias in imp.names:
        name = alias.name
        as_name = alias.asname or name
        yield name, as_name



"""
Exceptions for Command Parser

:author: Doug Skrypa
"""
# pylint: disable=W0231

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Optional, Collection, Mapping

from .utils import _parse_tree_target_repr

if TYPE_CHECKING:
    from .parameters import Parameter, BaseOption
    from .typing import ParamOrGroup
    from .parse_tree import PosNode, Word, Target

__all__ = [
    'CommandParserException',
    'ParserExit',
    'CommandDefinitionError',
    'ParameterDefinitionError',
    'AmbiguousShortForm',
    'AmbiguousParseTree',
    'UsageError',
    'ParamUsageError',
    'MultiParamUsageError',
    'AmbiguousCombo',
    'ParamConflict',
    'ParamsMissing',
    'BadArgument',
    'InvalidChoice',
    'MissingArgument',
    'NoSuchOption',
    'NoActiveContext',
]


class CommandParserException(Exception):
    """Base class for all other Command Parser exceptions"""

    code: int = 2

    def show(self):
        message = str(self)
        if message:
            print(message, file=sys.stderr)

    def exit(self):
        self.show()
        sys.exit(self.code)


class ParserExit(CommandParserException):
    """Exception used to exit with the given message and status code"""

    def __init__(self, message: str = None, code: int = None):
        self.code = code
        self.message = message

    def __str__(self):
        return self.message or ''


# region Developer Errors


class CommandDefinitionError(CommandParserException):
    """An error caused by providing invalid options for a Command, or an invalid combination of Parameters"""


class ParameterDefinitionError(CommandParserException):
    """An error caused by providing invalid options for a Parameter"""


class AmbiguousShortForm(ParameterDefinitionError):
    """
    Raised when a Parameter's short form contains multiple characters that would result in potentially ambiguous
    combinations with other Parameters' short forms.

    This will only be raised if ``config.ambiguous_short_combos`` is set to ``AmbiguousComboMode.STRICT``
    """

    def __init__(self, param_conflicts_map: Mapping[BaseOption, Collection[BaseOption]]):
        self.param_conflicts_map = param_conflicts_map

    def __str__(self) -> str:
        lines = []
        for param, conflicts in self.param_conflicts_map.items():
            param_str = param.format_usage(full=True, delim=' / ')
            conflicts_str = ', '.join(p.format_usage(full=True, delim=' / ') for p in conflicts)
            lines.append(f'Ambiguous short form for {param_str} - it conflicts with: {conflicts_str}')

        lines.sort()
        return '\n'.join(lines)


class AmbiguousParseTree(CommandDefinitionError):
    """Raised when a combination of parameters would result in ambiguous paths to take when parsing arguments"""

    def __init__(self, node: PosNode, target: Target, word: Word = None):
        self.node = node
        self.target = target
        self.word = word

    def __str__(self) -> str:
        node, word = self.node, self.word
        nt, st = _parse_tree_target_repr(node.target), _parse_tree_target_repr(self.target)
        if not word or word == node.word:
            return f'Conflicting targets for parse path={node.path_repr()}: {nt}, {st}'
        return f'Conflicting choices after parse path={node.parent.path_repr()}: {node.word}=>{nt}, {word}=>{st}'


# endregion

# region User Errors


class UsageError(CommandParserException):
    """Base exception for user errors"""

    message: str = None


class ParamUsageError(UsageError):
    """Error raised when a Parameter was not used correctly"""

    def __init__(self, param: Optional[ParamOrGroup], message: str = None):
        self.param = param
        self.usage_str = param.format_usage(full=True, delim=' / ') if param else ''
        if message:
            self.message = message

    def __str__(self) -> str:
        message = self.message or 'usage error'
        if self.param is None:
            return message
        else:
            return f'argument {self.usage_str}: {message}'


class MultiParamUsageError(UsageError):
    """Error raised when a combination of Parameters was not used correctly"""

    def __init__(self, params: Collection[ParamOrGroup], message: str = None):
        self.params = params
        self.usage_str = ', '.join(sorted(param.format_usage(full=True, delim=' / ') for param in params))
        if message:
            self.message = message

    def _usage_msg(self) -> str:
        if self.message:
            return f'{self.usage_str} ({self.message})'
        return self.usage_str

    def __str__(self) -> str:
        return f'usage error for the following combination of arguments: {self._usage_msg()}'


class AmbiguousCombo(MultiParamUsageError):
    """Error raised when an ambiguous combination of short options were provided"""

    def __init__(self, params: Collection[ParamOrGroup], combo: str, message: str = None):
        super().__init__(params, message)
        self.combo = combo

    def __str__(self) -> str:
        return (
            f'ambiguous option combo - part of argument={self.combo!r}'
            f' may match multiple parameters: {self._usage_msg()}'
        )


class ParamConflict(MultiParamUsageError):
    """Error raised when mutually exclusive Parameters were combined"""

    def __str__(self) -> str:
        return f'argument conflict - the following arguments cannot be combined: {self._usage_msg()}'


class ParamsMissing(UsageError):
    """Error raised when one or more required Parameters were not provided"""

    def __init__(self, params: Collection[ParamOrGroup], message: str = None):
        self.params = params
        self.usage_str = ', '.join(param.format_usage(full=True, delim=' / ') for param in params)
        if message:
            self.message = message

    def __str__(self) -> str:
        message = f' ({self.message})' if self.message else ''
        if not message:
            message = '; '.join(p.missing_hint for p in self.params if p.missing_hint)

        if len(self.params) > 1:
            prefix = 'arguments missing - the following arguments are required'
        else:
            prefix = 'argument missing - the following argument is required'
        return f'{prefix}: {self.usage_str}{message}'


class BadArgument(ParamUsageError):
    """Error raised when an invalid value is provided for a Parameter"""


class InvalidChoice(BadArgument):
    """Error raised when a value that does not match one of the pre-defined choices was provided for a Parameter"""

    def __init__(self, param: Optional[Parameter], invalid: Any, choices: Collection[Any]):
        if isinstance(invalid, Collection) and not isinstance(invalid, str):
            bad_str = 'choices: {}'.format(', '.join(map(repr, invalid)))
        else:
            bad_str = f'choice: {invalid!r}'
        choices_str = ', '.join(map(repr, choices))
        super().__init__(param, f'invalid {bad_str} (choose from: {choices_str})')


class MissingArgument(BadArgument):
    """Error raised when a value for a Parameter was not provided"""

    message = 'missing required argument value'


class NoSuchOption(UsageError):
    """Error raised when an option that was not defined as a Parameter was provided"""


class NoActiveContext(CommandParserException, RuntimeError):
    """Raised when attempting to perform an action that requires an active context while no context is active."""


# endregion

# region Internal Exceptions


class UnsupportedAction(CommandParserException):
    """Indicates that an attempted action cannot be completed.  Only used internally."""


class Backtrack(CommandParserException):
    """Raised when backtracking took place"""


# endregion



from __future__ import annotations

import keyword
import logging
from abc import ABC, abstractmethod
from ast import literal_eval, Attribute, Name, GeneratorExp, Subscript, DictComp, ListComp, SetComp
from dataclasses import dataclass, fields
from itertools import count
from typing import TYPE_CHECKING, Union, Optional, Iterator, Iterable, Type, TypeVar, Generic, List, Tuple

from cli_command_parser.compat import cached_property
from cli_command_parser.nargs import Nargs
from .argparse_ast import AC, ParserArg, ArgGroup, MutuallyExclusiveGroup, AstArgumentParser, Script
from .utils import collection_contents, unparse

if TYPE_CHECKING:
    from cli_command_parser.typing import OptStr
    from .argparse_ast import ArgCollection

__all__ = ['convert_script']
log = logging.getLogger(__name__)

C = TypeVar('C', bound='Converter')

RESERVED = set(keyword.kwlist) | set(getattr(keyword, 'softkwlist', ('_', 'case', 'match')))  # soft was added in 3.9


def convert_script(script: Script) -> str:
    return ScriptConverter(script).convert()


class Converter(ABC):
    converts: Type[AC] = None
    _ac_converter_map = {}

    def __init_subclass__(cls, converts: Type[AC] = None, **kwargs):
        super().__init_subclass__(**kwargs)
        if converts:
            cls.converts = converts
            cls._ac_converter_map[converts] = cls

    def __init__(self, ast_obj: Union[AC, Script], parent: Optional[Converter] = None):
        self.ast_obj = ast_obj
        self.parent = parent

    @classmethod
    def for_ast_callable(cls, ast_obj: Union[AC, Type[AC]]) -> Type[Converter]:
        if not isinstance(ast_obj, type):
            ast_obj = ast_obj.__class__
        try:
            return cls._ac_converter_map[ast_obj]
        except KeyError:
            pass
        for converts_cls, converter_cls in cls._ac_converter_map.items():
            if issubclass(ast_obj, converts_cls):
                return converter_cls
        raise TypeError(f'No Converter is registered for {ast_obj.__class__.__name__} objects')

    @classmethod
    def init_group(cls: Type[C], parent: CollectionConverter, ast_objs: List[AC]) -> ConverterGroup[C]:
        return ConverterGroup(parent, cls, [cls(ast_obj, parent) for ast_obj in ast_objs])

    def convert(self, indent: int = 0) -> str:
        return '\n'.join(self.format_lines(indent))

    @abstractmethod
    def format_lines(self, indent: int = 0) -> Iterator[str]:
        raise NotImplementedError


class ConverterGroup(Generic[C]):
    __slots__ = ('parent', 'member_type', 'members')

    def __init__(self, parent: CollectionConverter, member_type: Type[C], members: List[C]):
        self.parent = parent
        self.member_type = member_type
        self.members = members

    def __len__(self) -> int:
        return len(self.members)

    def __getitem__(self, index: int) -> C:
        return self.members[index]

    def __iter__(self) -> Iterator[C]:
        yield from self.members

    def format_all(self, indent: int = 0) -> Iterator[str]:
        for member in self.members:
            yield from member.format_lines(indent)


class ScriptConverter(Converter, converts=Script):
    def format_lines(self, indent: int = 0) -> Iterator[str]:
        # TODO: Filter to what is actually used
        yield (
            'from cli_command_parser import'
            ' Command, SubCommand, ParamGroup, Positional, Option, Flag, Counter, PassThru, main'
        )
        counter = count()
        for parser in self.ast_obj.parsers:
            yield from ParserConverter(parser, counter=counter).format_lines()


class CollectionConverter(Converter, ABC):
    ast_obj: ArgCollection
    parent: CollectionConverter | None
    _name_mode = None

    @cached_property
    def name_mode(self) -> str | None:
        return self._name_mode or (self.parent.name_mode if self.parent else None)

    @cached_property
    def grouped_children(self) -> List[ConverterGroup[ParamConverter | GroupConverter | Converter]]:
        return [self.for_ast_callable(cg_cls).init_group(self, cg) for cg_cls, cg in self.ast_obj.grouped_children()]

    def descendant_args(self) -> Iterator[ParamConverter]:
        for child_group in self.grouped_children:
            if not child_group.members:
                continue
            elif hasattr(child_group[0], 'descendant_args'):
                for child in child_group:
                    yield from child.descendant_args()
            elif isinstance(child_group[0], ParamConverter):
                yield from child_group

    def format_members(self, prefix: str, indent: int = 4) -> Iterator[str]:
        last = False
        for child_group in self.grouped_children:
            if last and child_group and issubclass(child_group.member_type, GroupConverter):
                yield ''
            yield from child_group.format_all(indent)
            last = bool(child_group)

        if not any(cg for cg in self.grouped_children):
            yield f'{prefix}    pass'


class ParserConverter(CollectionConverter, converts=AstArgumentParser):
    _auto_gen_disclaimer = '# This is an automatically generated name that should probably be updated'
    ast_obj: AstArgumentParser
    parent: ParserConverter | None

    def __init__(self, parser: AstArgumentParser, parent: ParserConverter = None, counter: count = None):
        super().__init__(parser, parent)
        self.counter = count() if counter is None else counter

    @cached_property
    def sub_parser_converters(self) -> List[ParserConverter]:
        return [self.__class__(sub_parser, self, self.counter) for sub_parser in self.ast_obj.sub_parsers]

    def descendant_args(self) -> Iterator[ParamConverter]:
        yield from super().descendant_args()
        for sp_converter in self.sub_parser_converters:
            yield from sp_converter.descendant_args()

    def format_lines(self, indent: int = 0) -> Iterator[str]:
        suffix = f'  {self._auto_gen_disclaimer}' if self.parent is None else ''
        # TODO: Add _init_command_ and/or main methods
        # TODO: If subparsers have no unique args, use action methods instead?
        yield '\n'
        yield f'class {self.name}({self._get_args()}):{suffix}'
        yield from self.format_members('')
        for sp_converter in self.sub_parser_converters:
            yield from sp_converter.format_lines()

    def _get_args(self) -> str:
        kwargs = self.ast_obj.init_func_kwargs.copy()
        # log.debug(f'Processing args for {kwargs}')
        kwargs['option_name_mode'] = self._name_mode
        if self.is_sub_parser:
            key, value = self._choices
            if key:
                kwargs[key] = value
        elif 'add_help' in kwargs and literal_eval_or_none(kwargs['add_help']) is True:
            kwargs.pop('add_help')

        cmd_args = CommandArgs.from_kwargs(**kwargs)
        return cmd_args.to_str(self.parent.name if self.parent else 'Command')

    @cached_property
    def is_sub_parser(self) -> bool:
        return self.parent is not None

    # region Name / CLI Choices

    @cached_property
    def name(self) -> str:
        return self._custom_name or f'Command{next(self.counter)}'

    @cached_property
    def _custom_name(self) -> OptStr:
        if not self.is_sub_parser:
            return None
        name = literal_eval_or_none(self.ast_obj.init_func_kwargs.get('name'))
        if not name or not name[0].isalpha():
            return None
        return name.title().replace(' ', '').replace('_', '').replace('-', '')

    @cached_property
    def _choices(self) -> Tuple[OptStr, OptStr]:
        if not self.is_sub_parser:
            return None, None
        name = self.ast_obj.init_func_kwargs.get('name')
        aliases = self.ast_obj.init_func_raw_kwargs.get('aliases')
        if not aliases:
            if name and (not self._custom_name or '-' in name or ' ' in name):
                return 'choice', name
            return None, None
        elif isinstance(aliases, (Attribute, Name, Subscript, GeneratorExp, DictComp, ListComp, SetComp)):
            value = unparse(aliases)  # noqa
            if name:
                return 'choices', f'({name}, *{value})'
            return 'choices', (f'tuple{value}' if isinstance(aliases, GeneratorExp) else value)
        else:
            parsed = collection_contents(aliases)
            values = [name, *parsed] if name else parsed
            if len(values) == 1:
                return 'choice', values[0]
            elif values:
                return 'choices', f'({", ".join(values)})'
            return None, None

    # endregion

    # region Member-Related Properties

    @cached_property
    def name_mode(self) -> str | None:
        return self._name_mode or (self.parent.name_mode if self.parent else None)

    @cached_property
    def _name_mode(self) -> str | None:
        if self.parent and self.parent._name_mode:
            return None
        name_modes = {pc._name_mode for pc in self.descendant_args() if pc.is_option and '_' in pc.attr_name}
        return next(iter(name_modes)) if len(name_modes) == 1 else None

    # endregion


class GroupConverter(CollectionConverter, converts=ArgGroup):
    ast_obj: ArgGroup

    def format_lines(self, indent: int = 4) -> Iterator[str]:
        prefix = ' ' * indent
        yield f'{prefix}with ParamGroup({self._get_args()}):'
        yield from self.format_members(prefix, indent + 4)

    def _get_args(self) -> str:
        # log.debug(f'Processing args for {self.ast_obj._init_func_bound}')
        description = self.ast_obj.init_func_kwargs.get('description')
        title = self.ast_obj.init_func_kwargs.get('title')
        if title:
            title_str = literal_eval(title)
            if title_str.lower().endswith(' options'):
                if description:
                    title = repr(title_str[:-7].rstrip())
                else:
                    description, title = title, None

        args = [title] if title else []
        if description:
            args.append(f'description={description}')
        if isinstance(self.ast_obj, MutuallyExclusiveGroup):
            args.append('mutually_exclusive=True')
        return ', '.join(args)


class ParamConverter(Converter, converts=ParserArg):
    ast_obj: ParserArg
    parent: CollectionConverter | None
    _counter = count()

    def __init__(self, arg: ParserArg, parent: CollectionConverter, num: int):
        super().__init__(arg, parent)
        self.num = num

    def __eq__(self, other: ParamConverter) -> bool:
        return self.ast_obj == other.ast_obj and self.num == other.num

    def __lt__(self, other: ParamConverter) -> bool:
        if self.is_positional and not other.is_positional:
            return True
        if self.is_pass_thru and not other.is_pass_thru:
            return False
        return self.num < other.num

    @classmethod
    def init_group(cls, parent: CollectionConverter, args: List[ParserArg]) -> ParamConverterGroup:
        return ParamConverterGroup(parent, cls, [cls(arg, parent, i) for i, arg in enumerate(args)])

    def format_lines(self, indent: int = 4) -> Iterator[str]:
        yield self.format(indent)

    def format(self, indent: int = 4) -> str:
        param_cls, args_obj = self.get_cls_and_kwargs()
        arg_str = args_obj.to_str(*self.get_pos_args())
        return f'{" " * indent}{self.attr_name} = {param_cls}({arg_str})'

    # region Naming

    @cached_property
    def attr_name(self) -> str:
        return self._attr_name.replace('-', '_')

    @cached_property
    def name_mode(self) -> str | None:
        return None if self.parent.name_mode else self._name_mode

    @cached_property
    def _name_mode(self) -> str | None:
        if not self.use_auto_long_opt_str:
            return None
        return "'_'" if '_' in self._attr_name else None

    @cached_property
    def _attr_name(self) -> str:
        return next(name for name in self._attr_name_candidates() if name not in RESERVED)

    def _attr_name_candidates(self) -> Iterator[str]:
        long, short, plain = self._grouped_opt_strs
        if self.is_positional or self.is_pass_thru:
            yield from plain
        if self.is_option or self.is_pass_thru:
            for group in (long, short):
                for opt in group:
                    opt = opt.lstrip('-')
                    if opt:
                        yield opt
        while True:
            yield f'param_{next(self._counter)}'

    # endregion

    # region Arg Processing

    @cached_property
    def cmd_option_strs(self) -> List[str]:
        if not self.is_option:
            return []
        long, short, plain = self._grouped_opt_strs
        return short if self.use_auto_long_opt_str else (long + short)

    @cached_property
    def use_auto_long_opt_str(self) -> bool:
        if not self.is_option:
            return False
        long, short, plain = self._grouped_opt_strs
        return len(long) == 1 and long[0][2:] == self._attr_name

    def get_pos_args(self) -> Iterable[str]:
        return (repr(arg) for arg in self.cmd_option_strs)

    def get_cls_and_kwargs(self) -> Tuple[str, BaseArgs]:
        kwargs = self.ast_obj.init_func_kwargs.copy()
        help_arg = kwargs.get('help')
        if help_arg and help_arg in self.ast_obj.get_tracked_refs('argparse', 'SUPPRESS', ()):
            kwargs.update({'hide': 'True', 'help': None})

        if self.is_pass_thru:
            return 'PassThru', PassThruArgs.from_kwargs(**kwargs)

        action = kwargs.pop('action', None)
        if action:
            action = literal_eval(action)

        if self.is_positional:
            if action and action not in ('store', 'append'):
                raise ConversionError(f'{self.ast_obj}: action={action!r} is not supported for Positional parameters')
            return 'Positional', ParamArgs.init_positional(action, **kwargs)
        elif self.is_option:
            kwargs['name_mode'] = self.name_mode
            if not action and 'const' in kwargs:
                action = 'append_const' if 'nargs' in kwargs else 'store_const'
            if action:
                if action in ('store_true', 'store_false', 'store_const', 'append_const'):
                    return 'Flag', FlagArgs.init_flag(action, **kwargs)
                elif action == 'count':
                    return 'Counter', FlagArgs.init_counter(**kwargs)
                elif action not in ('store', 'append'):
                    raise ConversionError(f'{self.ast_obj}: action={action!r} is not supported for Option parameters')

            return 'Option', OptionArgs.init_option(self.ast_obj, action, **kwargs)

        raise ConversionError(f'Unable to determine a suitable Parameter type for {self.ast_obj!r}')

    # endregion

    # region High Level Param Type

    @cached_property
    def is_pass_thru(self) -> bool:
        nargs = self.ast_obj.init_func_kwargs.get('nargs')
        if not nargs:
            return False
        return nargs in self.ast_obj.get_tracked_refs('argparse', 'REMAINDER', ())

    @cached_property
    def is_positional(self) -> bool:
        long, short, plain = self._grouped_opt_strs
        return plain and not long and not short

    @cached_property
    def is_option(self) -> bool:
        long, short, plain = self._grouped_opt_strs
        return (long or short) and not plain

    # endregion

    @cached_property
    def _grouped_opt_strs(self) -> Tuple[List[str], List[str], List[str]]:
        option_strs = (literal_eval(opt) for opt in self.ast_obj.init_func_args)
        long, short, plain = [], [], []
        for opt in option_strs:
            if opt.startswith('--'):
                long.append(opt)
            elif opt.startswith('-'):
                short.append(opt)
            else:
                plain.append(opt)
        return long, short, plain


class ParamConverterGroup(ConverterGroup[ParamConverter]):
    __slots__ = ()

    def __bool__(self) -> bool:
        return bool(self.members) or bool(getattr(self.parent.ast_obj, 'sub_parsers', None))

    def format_all(self, indent: int = 4) -> Iterator[str]:
        positionals, others = [], []
        i_converters = iter(sorted(self.members))
        for converter in i_converters:
            if converter.is_positional:
                positionals.append(converter)
            else:
                others.append(converter)
                others.extend(i_converters)

        for positional in positionals:
            yield from positional.format_lines(indent)

        sub_parsers = getattr(self.parent.ast_obj, 'sub_parsers', None)
        if sub_parsers:
            log.debug(f'Found sub_parsers={sub_parsers}')
            try:
                name = literal_eval(sub_parsers[0].init_func_kwargs['dest']).replace('-', '_')
            except (KeyError, ValueError):
                name = 'sub_cmd'
            else:
                if name in RESERVED:
                    name = 'sub_cmd'
            yield f'{" " * indent}{name} = SubCommand()'

        for other in others:
            yield from other.format_lines(indent)


# region Arg Containers


@dataclass
class BaseArgs:
    help: OptStr = None

    def _to_str(self, args: Tuple[str, ...], end_fields: List[str]) -> str:
        skip = set(end_fields)
        keys = [f.name for f in fields(self) if f.name not in skip] + end_fields
        kv_iter = ((key, getattr(self, key)) for key in keys)
        return ', '.join((*args, *(f'{key}={val}' for key, val in kv_iter if val is not None)))

    def to_str(self, *args: str) -> str:
        return self._to_str(args, ['help'])

    @classmethod
    def from_kwargs(cls, **kwargs):
        keys = set(f.name for f in fields(cls)).intersection(kwargs)
        filtered = {key: kwargs[key] for key in keys}
        help_str = filtered.get('help')
        if help_str:
            # log.debug(f'Processing {help_str=}')
            try:
                help_str = literal_eval(help_str)
            except ValueError:  # likely an f-string
                pass
            else:
                if help_str.endswith('(default: %(default)s)'):
                    help_str = help_str[:-22].rstrip()
                filtered['help'] = repr(help_str) if help_str else None

        return cls(**filtered)


@dataclass
class CommandArgs(BaseArgs):
    choice: OptStr = None
    choices: OptStr = None
    prog: OptStr = None
    usage: OptStr = None
    description: OptStr = None
    epilog: OptStr = None
    option_name_mode: OptStr = None
    add_help: OptStr = None
    docs_url: OptStr = None
    email: OptStr = None


# endregion


# region ParserArg Arg Containers


@dataclass
class ParamBaseArgs(BaseArgs):
    name: OptStr = None
    default: OptStr = None
    required: OptStr = None
    metavar: OptStr = None
    hide: OptStr = None

    def to_str(self, *args: str) -> str:
        return self._to_str(args, ['hide', 'help'])


@dataclass
class PassThruArgs(ParamBaseArgs):
    pass


@dataclass
class ParamArgs(ParamBaseArgs):
    action: OptStr = None
    type: OptStr = None
    nargs: OptStr = None
    choices: OptStr = None

    @classmethod
    def init_positional(cls, action: OptStr = None, nargs: OptStr = None, **kwargs):
        if nargs is not None:
            parsed = literal_eval_or_none(nargs)
            if parsed is not None:
                nargs_obj = Nargs(parsed)
                if action in ('store', None) and nargs_obj == 1:
                    action = nargs = None
            else:
                nargs_obj = None
        else:
            nargs_obj = Nargs(1)

        if nargs_obj is not None and action == 'append' and nargs_obj != Nargs(1):
            action = None
        return cls.from_kwargs(action=action, nargs=nargs, **kwargs)


@dataclass
class OptionArgs(ParamArgs):
    name_mode: OptStr = None

    @classmethod
    def init_option(cls, arg: ParserArg, action: OptStr = None, nargs: OptStr = None, const: OptStr = None, **kwargs):
        if const:
            log.warning(f'{arg}: ignoring const={const!r} - it is only supported for Flag and Counter parameters')

        if nargs == "'*'":
            nargs = "'+'"
        if action == 'append':
            if not nargs:
                log.debug(f"{arg}: using default nargs='+' because action={action!r} and no nargs value was provided")
                nargs = "'+'"
            action = None
        elif action == 'store':
            if nargs == '1':
                nargs = None
            action = None

        return cls.from_kwargs(action=repr(action) if action else None, nargs=nargs, **kwargs)


@dataclass
class FlagArgs(OptionArgs):
    const: OptStr = None

    @classmethod
    def init_flag(cls, action: str, const: OptStr = None, default: OptStr = None, **kwargs):
        values = {'store_true': ('True', 'False'), 'store_false': ('False', 'True')}
        try:
            value, opposite = values[action]
        except KeyError:
            if action == 'store_const':
                action = None
        else:
            if default == opposite:
                default = None
            action = None
            const = value if default else None

        kwargs['type'] = kwargs['nargs'] = None
        if action:
            action = repr(action)
        return cls.from_kwargs(action=action, const=const, default=default, **kwargs)

    @classmethod
    def init_counter(cls, const: OptStr = None, default: OptStr = None, **kwargs):
        kwargs['type'] = kwargs['nargs'] = kwargs['action'] = None
        kwargs['const'] = None if const == '1' else const
        kwargs['default'] = None if default == '0' else default
        return cls.from_kwargs(**kwargs)


# endregion


def literal_eval_or_none(expr: str) -> str | None:
    try:
        return literal_eval(expr)
    except ValueError:
        return None


class ConversionError(Exception):
    pass



"""
Utilities for working with terminals, strings, and Enums.

:author: Doug Skrypa
"""

from __future__ import annotations

from enum import Flag, EnumMeta
from shutil import get_terminal_size
from time import monotonic
from typing import Any, Callable, TypeVar, List

try:
    from enum import CONFORM
except ImportError:
    CONFORM = None

FlagEnum = TypeVar('FlagEnum', bound='FixedFlag')
_NotSet = object()


# region Text Processing / Formatting


def camel_to_snake_case(text: str, delim: str = '_') -> str:
    return ''.join(f'{delim}{c}' if i and c.isupper() else c for i, c in enumerate(text)).lower()


def short_repr(obj: Any, max_len: int = 100, sep: str = '...', func: Callable[[Any], str] = repr) -> str:
    obj_repr = func(obj)
    if len(obj_repr) > max_len:
        part_len = (max_len - len(sep)) // 2
        return '{}{}{}'.format(obj_repr[:part_len], sep, obj_repr[-part_len:])
    return obj_repr


def _parse_tree_target_repr(target) -> str:
    try:
        return target.__name__
    except AttributeError:
        return repr(target)


# endregion


class MissingMixin:
    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str):
            try:
                return cls._member_map_[value.upper()]  # noqa
            except KeyError:
                pass
        return super()._missing_(value)  # noqa


class FixedFlagMeta(EnumMeta):
    """
    This metaclass is only used to maintain the same behavior for Flag pseudo-members between 3.7-3.10 and 3.11 when
    the behavior was parameterized via the ``boundary`` parameter.
    """

    def __new__(mcs, *args, **kwargs):
        if CONFORM:
            kwargs['boundary'] = CONFORM
        return super().__new__(mcs, *args, **kwargs)


class FixedFlag(Flag, metaclass=FixedFlagMeta):
    """Extends Flag to work around breaking changes in 3.11 for repr, missing, and pseudo-members."""

    def __repr__(self) -> str:
        # In 3.11, this needs to be declared in the parent of a Flag that actually has members - it breaks if it is
        # defined in a mixin or the class with members.
        names = '|'.join(part._name_ for part in self._decompose())  # noqa
        return f'<{self.__class__.__name__}:{names}>'

    @classmethod
    def _missing_(cls, value) -> FlagEnum:
        if isinstance(value, str):
            if value.startswith(('!', '~')):
                invert = True
                value = value[1:]
            else:
                invert = False

            try:
                member = cls._missing_str(value)
            except KeyError:
                expected = ', '.join(cls._member_map_)
                raise ValueError(f'Invalid {cls.__name__} value={value!r} - expected one of {expected}') from None
            else:
                return ~member if invert else member  # pylint: disable=E1130
        elif not isinstance(value, int):
            raise TypeError(f'Unexpected type={value.__class__.__name__} for a {cls.__name__}')

        return super()._missing_(value)

    @classmethod
    def _missing_str(cls, value: str) -> FlagEnum:
        try:
            return cls._member_map_[value.upper()]  # noqa
        except KeyError:
            pass
        if '|' in value:
            tmp = cls(0)
            for part in map(str.strip, value.split('|')):
                if not part:
                    continue
                try:
                    tmp |= cls._member_map_[part.upper()]
                except KeyError:
                    break
            else:
                if tmp._value_ != 0:
                    return tmp

        raise KeyError

    def _decompose(self) -> List[FlagEnum]:
        if self._name_ is None or '|' in self._name_:  # | check is for 3.11 where pseudo-members are assigned names
            val = self._value_
            members = ((mem, mem._value_) for mem in self.__class__)
            return sorted(mem for mem, mem_val in members if mem_val & val == mem_val)
        return [self]

    def __lt__(self, other: FlagEnum) -> bool:
        return self._value_ < other._value_


class Terminal:  # pylint: disable=R0903
    __slots__ = ('_cache_time', '_last_time', '_width')

    def __init__(self, cache_time: float = 1):
        self._cache_time = cache_time
        self._last_time = 0
        self._width = 80

    @property
    def width(self) -> int:
        if monotonic() - self._last_time > self._cache_time:
            self._width = get_terminal_size()[0]
            self._last_time = monotonic()
        return self._width



"""
Program metadata introspection for use in usage, help text, and documentation.

:author: Doug Skrypa
"""
# pylint: disable=R0801

from __future__ import annotations

from inspect import getmodule
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Type, Optional, Union, Tuple, Dict
from urllib.parse import urlparse

from .context import ctx, NoActiveContext

if TYPE_CHECKING:
    from .typing import Bool, CommandType, OptStr

__all__ = ['ProgramMetadata']

DEFAULT_FILE_NAME: str = 'UNKNOWN'


class Metadata:
    __slots__ = ('default', 'name')

    def __init__(self, default):
        self.default = default

    def __set_name__(self, owner: Type[ProgramMetadata], name: str):
        self.name = name
        owner._fields.add(name)

    def __get__(self, instance: Optional[ProgramMetadata], owner: Type[ProgramMetadata]):
        if instance is None:
            return self
        try:
            return instance.__dict__[self.name]
        except KeyError:
            pass
        try:
            return getattr(instance.parent, self.name)
        except AttributeError:  # parent is None
            return self.default

    def __set__(self, instance: ProgramMetadata, value: Union[str, Path, None]):
        if value is not None:
            instance.__dict__[self.name] = value

    def __repr__(self) -> str:
        return f'Metadata(default={self.default!r})'


class ProgramMetadata:
    _fields = {'parent'}
    parent: Optional[ProgramMetadata] = None
    path: Path = Metadata(None)
    package: str = Metadata(None)
    module: str = Metadata(None)
    command: str = Metadata(None)
    prog: str = Metadata(None)
    prog_from_sys_argv: bool = Metadata(None)
    url: str = Metadata(None)
    docs_url: str = Metadata(None)
    email: str = Metadata(None)
    version: str = Metadata('')
    usage: str = Metadata(None)
    description: str = Metadata(None)
    epilog: str = Metadata(None)
    doc_name: str = Metadata(None)
    doc_str: str = Metadata('')
    pkg_doc_str: str = Metadata('')  # Set by :func:`~.documentation.load_commands` to capture package docstrings

    def __init__(self, **kwargs):
        fields = self._fields
        for key, val in kwargs.items():
            if key in fields:
                setattr(self, key, val)
            else:
                # The number of times one or more invalid options will be provided is extremely low compared to how
                # often this exception will not need to be raised, so the re-iteration over kwargs is acceptable.
                # This also avoids creating the `bad` dict that would otherwise be thrown away on 99.9% of init calls.
                bad = ', '.join(sorted(key for key in kwargs if key not in fields))
                raise TypeError(f'Invalid arguments for {self.__class__.__name__}: {bad}')

    @classmethod
    def for_command(  # pylint: disable=R0914
        cls,
        command: CommandType,
        *,
        parent: ProgramMetadata = None,
        path: Path = None,
        prog: str = None,
        url: str = None,
        docs_url: str = None,
        email: str = None,
        version: str = None,
        usage: str = None,
        description: str = None,
        epilog: str = None,
        doc_name: str = None,
        no_sys_argv: Bool = None,
    ) -> ProgramMetadata:
        path, g = _path_and_globals(command, path)
        if command.__module__ != 'cli_command_parser.commands':
            # Prevent inheritors from getting docstrings from the base Command
            doc_str = g.get('__doc__')
            doc = command.__doc__
        else:
            doc = doc_str = None

        prog, prog_from_sys_argv = _prog(prog, path, parent, no_sys_argv)
        return cls(
            parent=parent,
            path=path,
            package=g.get('__package__'),
            module=g.get('__module__'),
            command=command.__qualname__,
            prog=prog,
            prog_from_sys_argv=prog_from_sys_argv,
            url=url or g.get('__url__'),
            docs_url=docs_url or _docs_url_from_repo_url(url) or _docs_url_from_repo_url(g.get('__url__')),
            email=email or g.get('__author_email__'),
            version=version or g.get('__version__'),
            usage=usage,
            description=_description(description, doc),
            epilog=epilog,
            doc_name=_doc_name(doc_name, path, prog),
            doc_str=doc_str,
        )

    def __repr__(self) -> str:
        return _repr(self)

    def format_epilog(self, extended: Bool = True) -> str:
        parts = [self.epilog] if self.epilog else []
        if parts and not extended:
            return parts[0]

        version = self.version
        if version:
            version = f' [ver. {version}]'
        if self.email:
            parts.append(f'Report {self.prog}{version} bugs to {self.email}')
        url = self.docs_url or self.url
        if url:
            parts.append(f'Online documentation: {url}')
        return '\n\n'.join(parts)

    def get_doc_str(self, strip: bool = True) -> OptStr:
        doc_str = self.pkg_doc_str
        if doc_str and strip:
            doc_str = doc_str.strip()
        if not doc_str:
            doc_str = self.doc_str
            if doc_str and strip:
                doc_str = doc_str.strip()
        return doc_str


def _repr(obj, indent=0) -> str:
    if not isinstance(obj, ProgramMetadata):
        return repr(obj)

    field_dict = {field: getattr(obj, field) for field in obj._fields}
    prev_str = ' ' * indent
    indent += 4
    indent_str = ' ' * indent
    fields_str = '\n'.join(f'{indent_str}{k}={_repr(v, indent)},' for k, v in field_dict.items())
    return f'<{obj.__class__.__name__}(\n{fields_str}\n{prev_str})>'


def _prog(prog: OptStr, cmd_path: Path, parent: Optional[ProgramMetadata], no_sys_argv: Bool) -> Tuple[OptStr, bool]:
    # TODO: Attempt to detect the name to use via importlib.metadata.entry_points?  3.8+, with return value changes
    #  after 3.9
    if prog:
        return prog, False
    if no_sys_argv is None:
        try:
            no_sys_argv = not ctx.allow_argv_prog
        except NoActiveContext:
            no_sys_argv = False

    if parent and parent.prog != parent.path.name and (not no_sys_argv or not parent.prog_from_sys_argv):
        return parent.prog, parent.prog_from_sys_argv
    elif not no_sys_argv:
        try:
            ctx_prog = ctx.prog
        except NoActiveContext:
            ctx_prog = None

        if ctx_prog:
            path = Path(ctx_prog)
            # Windows allows invocation without .exe - assume a file with an extension is a match
            if path.exists() or next(path.parent.glob(f'{path.name}.???'), None) is not None:
                return path.name, True

    return cmd_path.name, False


def _path_and_globals(command: CommandType, path: Path = None) -> Tuple[Path, Dict[str, Any]]:
    module = getmodule(command)
    if path is None:
        try:
            path = Path(module.__file__).resolve()
        except AttributeError:  # module is None
            path = Path.cwd().joinpath(DEFAULT_FILE_NAME)

    if module is None:
        return path, {}

    return path, module.__dict__


def _description(description: Optional[str], doc: Optional[str]) -> Optional[str]:
    if description:
        return description
    elif doc:
        doc = dedent(doc).lstrip()
        if doc.strip():  # avoid space-only doc, but let possibly intentional trailing spaces / newlines to persist
            return doc
    return None


def _docs_url_from_repo_url(repo_url: Optional[str]) -> Optional[str]:
    if not repo_url:
        return None

    parsed = urlparse(repo_url)
    if parsed.scheme == 'https' and parsed.hostname == 'github.com':
        try:
            user, repo = parsed.path[1:].split('/', 1)
        except ValueError:
            pass
        else:
            return f'https://{user}.github.io/{repo}/'

    return None


def _doc_name(doc_name: Optional[str], path: Path, prog: str) -> str:
    if doc_name:
        return doc_name
    return Path(prog).stem



from .argparse_ast import Script, AstCallable, ArgCollection, AstArgumentParser, SubParser, AddVisitedChild, visit_func
from .argparse_ast import ArgGroup, MutuallyExclusiveGroup
from .command_builder import Converter, convert_script



"""
Utilities for extracting types from annotations.

:author: Doug Skrypa
"""

from __future__ import annotations

from collections.abc import Collection, Iterable
from inspect import isclass
from typing import Union, Optional, get_type_hints

try:
    from typing import get_origin, get_args as _get_args  # pylint: disable=C0412
except ImportError:  # Added in 3.8; the versions from 3.8 are copied here
    from .compat import get_origin, _get_args

try:
    from types import NoneType
except ImportError:  # Added in 3.10
    NoneType = type(None)

__all__ = ['get_descriptor_value_type']


def get_descriptor_value_type(command_cls: type, attr: str) -> Optional[type]:
    try:
        annotation = get_type_hints(command_cls)[attr]
    except (KeyError, NameError):  # KeyError due to attr missing; NameError for forward references
        return None

    return get_annotation_value_type(annotation)


def get_annotation_value_type(annotation) -> Optional[type]:
    origin = get_origin(annotation)
    # Note on get_origin return values:
    # get_origin(List[str]) -> list
    # get_origin(List) -> list
    # get_origin(list[str]) -> list
    # get_origin(list) -> None
    if origin is None and isinstance(annotation, type):
        return annotation
    elif isclass(origin) and issubclass(origin, (Collection, Iterable)):
        return _type_from_collection(origin, annotation)
    elif origin is Union:
        return _type_from_union(annotation)
    return None


def get_args(annotation) -> tuple:
    """
    Wrapper around :func:`python:typing.get_args` for 3.7~8 compatibility, to make it behave more like it does in 3.9+
    """
    if getattr(annotation, '_special', False):  # 3.7-3.8 generic collection alias with no content types
        return ()
    return _get_args(annotation)


def _type_from_union(annotation) -> Optional[type]:
    args = get_args(annotation)
    # Note: Unions of a single argument return the argument; i.e., Union[T] returns T, so the len can never be 1
    if len(args) == 2 and NoneType in args:
        arg = args[0] if args[1] is NoneType else args[1]
    else:
        return None

    origin = get_origin(arg)
    if origin is None and isinstance(arg, type):
        return arg
    elif isclass(origin) and issubclass(origin, (Collection, Iterable)):
        return _type_from_collection(origin, arg)
    else:
        return None


def _type_from_collection(origin, annotation) -> Optional[type]:
    args = get_args(annotation)
    try:
        annotation = args[0]
    except IndexError:  # The annotation was a collection with no content types specified
        return origin

    n_args = len(args)
    if n_args > 1 and (origin is not tuple or not ((n_args == 2 and args[1] is Ellipsis) or len(set(args)) == 1)):
        return None

    origin = get_origin(annotation)
    if origin is None and isinstance(annotation, type):
        return annotation
    elif origin is Union:
        return _type_from_union(annotation)
    else:
        return None



"""
Typing helpers to make it easier to use Signature.from_callable for automatic conversion of positional to keyword args.
"""

from argparse import ArgumentParser as _ArgumentParser, _SubParsersAction  # noqa

__all__ = ['ArgumentParser', 'SubParsersAction']


class ArgumentParser(_ArgumentParser):
    def add_argument_group(
        self, title=None, description=None, *, prefix_chars=None, argument_default=None, conflict_handler=None
    ):
        kwargs = {k: v for k, v in locals().items() if k not in {'self', '__class__'} and v is not None}
        return super().add_argument_group(**kwargs)

    def add_mutually_exclusive_group(self, *, required=False):
        return super().add_mutually_exclusive_group(required=required)

    # fmt: off
    def add_subparsers(
        self, *, title=None, description=None, prog=None, dest=None, help=None,  # noqa
        action=None, option_string=None, required=None, metavar=None,
    ):
        kwargs = {k: v for k, v in locals().items() if k not in {'self', '__class__'} and v is not None}
        return super().add_subparsers(**kwargs)
    # fmt: on


class SubParsersAction(_SubParsersAction):
    def add_parser(self, name, *, aliases=(), description=None, prog=None, help=None):  # noqa
        return super().add_parser(name, aliases=aliases, description=description, prog=prog, help=help)



"""
The CommandParameters class in this module is used to process all of the attributes of a given :class:`.Command` when
it is defined to collect its :class:`.Parameter` / :class:`.ParamGroup` members.  It also checks for conflicts between
parameter definitions.

It has some involvement in the parsing process for :class:`.BaseOption` parameters.

:author: Doug Skrypa
"""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Optional, Collection, Iterator, List, Dict, Set, Tuple

try:
    from functools import cached_property
except ImportError:
    from .compat import cached_property

from .actions import help_action
from .config import CommandConfig, AmbiguousComboMode
from .exceptions import CommandDefinitionError, ParameterDefinitionError, ParamsMissing
from .exceptions import AmbiguousShortForm, AmbiguousCombo
from .parameters.base import ParamBase, Parameter, BaseOption, BasePositional
from .parameters import SubCommand, PassThru, ActionFlag, ParamGroup, Action, Option

if TYPE_CHECKING:
    from .context import Context
    from .formatting.commands import CommandHelpFormatter
    from .typing import CommandCls

__all__ = ['CommandParameters']

OptionMap = Dict[str, BaseOption]
ActionFlags = List[ActionFlag]


class CommandParameters:
    command: CommandCls                                  #: The Command associated with this CommandParameters object
    formatter: CommandHelpFormatter                      #: The formatter used for this Command's help text
    command_parent: Optional[CommandCls] = None          #: The parent Command, if any
    parent: Optional[CommandParameters] = None           #: The parent Command's CommandParameters
    action: Optional[Action] = None                      #: An Action Parameter, if specified
    _pass_thru: Optional[PassThru] = None                #: A PassThru Parameter, if specified
    sub_command: Optional[SubCommand] = None             #: A SubCommand Parameter, if specified
    action_flags: ActionFlags                            #: List of action flags
    split_action_flags: Tuple[ActionFlags, ActionFlags]  #: Action flags split by before/after main
    options: List[BaseOption]                            #: List of optional Parameters
    combo_option_map: OptionMap                          #: Mapping of {short opt: Parameter} (no dash characters)
    groups: List[ParamGroup]                             #: List of ParamGroup objects
    positionals: List[BasePositional]                    #: List of positional Parameters
    option_map: OptionMap                                #: Mapping of {--opt / -opt: Parameter}

    def __init__(self, command: CommandCls, command_parent: Optional[CommandCls], config: CommandConfig):
        self.command = command
        if command_parent is not None:
            self.command_parent = command_parent
            self.parent = command_parent.__class__.params(command_parent)

        self.config = config
        self._process_parameters()

    def __repr__(self) -> str:
        positionals = len(self.positionals)
        options = len(self.options)
        cls_name = self.__class__.__name__
        return f'<{cls_name}[command={self.command.__name__}, positionals={positionals!r}, options={options!r}]>'

    @property
    def pass_thru(self) -> Optional[PassThru]:
        if self._pass_thru:
            return self._pass_thru
        elif self.parent:
            return self.parent.pass_thru
        return None

    @cached_property
    def always_available_action_flags(self) -> Tuple[ActionFlag, ...]:
        """
        The :paramref:`.ActionFlag.before_main` :class:`.ActionFlag` actions that are always available, even if parsing
        failed (such as the one for ``--help``).
        """
        return tuple(af for af in self.action_flags if af.always_available)

    @cached_property
    def formatter(self) -> CommandHelpFormatter:
        from .formatting.commands import CommandHelpFormatter

        if self.config is None:
            formatter_factory = CommandHelpFormatter
        else:
            formatter_factory = self.config.command_formatter or CommandHelpFormatter
        formatter = formatter_factory(self.command, self)
        formatter.maybe_add_option(self._pass_thru)
        formatter.maybe_add_positionals(self.positionals)
        formatter.maybe_add_options(self.options)
        formatter.maybe_add_groups(self.groups)
        return formatter

    @cached_property
    def _classified_combo_options(self) -> Tuple[Dict[str, BaseOption], Dict[str, BaseOption]]:
        multi_char_combos = {}
        single_char_combos = {}
        for combo, param in self.combo_option_map.items():
            if len(combo) == 1:
                single_char_combos[combo] = param
            else:
                multi_char_combos[combo] = param
        return single_char_combos, multi_char_combos

    @cached_property
    def _potentially_ambiguous_combo_options(self) -> Dict[str, Tuple[BaseOption, Dict[str, BaseOption]]]:
        single_char_combos, multi_char_combos = self._classified_combo_options
        if not multi_char_combos:
            return {}

        ambiguous_combo_options = {}
        for combo, param in multi_char_combos.items():
            singles = {c: single_char_combos[c] for c in combo if c in single_char_combos}
            if singles:
                ambiguous_combo_options[combo] = (param, singles)

        return ambiguous_combo_options

    @property
    def _has_help(self) -> bool:
        return help_action in self.always_available_action_flags or (self.parent and self.parent._has_help)

    # region Initialization

    def _process_parameters(self):
        """
        Process all of the :class:`.Parameter` / :class:`.ParamGroup` members in the associated :class:`.Command` class.
        """
        name_param_map = {}  # Allow subclasses to override names, but not within a given command
        positionals = []
        options = []
        groups = set()

        for attr, param in self.command.__dict__.items():
            if attr.startswith('__') or not isinstance(param, ParamBase):  # Name mangled Parameters are still processed
                continue

            name = param.name
            try:
                other_attr, other_param = name_param_map[name]
            except KeyError:
                name_param_map[name] = (attr, param)
            else:
                raise CommandDefinitionError(
                    'Name conflict - multiple parameters within a Command cannot have the same name - conflicting'
                    f' params: {other_attr}={other_param}, {attr}={param}'
                )

            if isinstance(param, BasePositional):
                positionals.append(param)
            elif isinstance(param, BaseOption):
                options.append(param)
            elif isinstance(param, ParamGroup):
                # Groups will only be discovered here when defined with `as` - ex: `with ParamGroup(...) as foo:`
                groups.add(param)
            elif isinstance(param, PassThru):
                if self.pass_thru:
                    raise CommandDefinitionError(
                        f'Invalid PassThru param={param!r} - it cannot follow another PassThru param'
                    )
                self._pass_thru = param
            else:
                raise CommandDefinitionError(
                    f'Unexpected type={param.__class__} for param={param!r} - custom parameters must extend'
                    ' BasePositional, BaseOption, or ParamGroup'
                )

            if param.group:
                _find_groups(groups, param)

        if self.config and self.config.add_help and (not self.parent or not self.parent._has_help):
            options.append(help_action)

        self._process_positionals(positionals)
        self._process_options(options)
        self._process_groups(groups)

    def _process_groups(self, groups: Set[ParamGroup]):
        if self.parent:
            _groups, groups = groups, self.parent.groups.copy()
            groups.extend(_groups)

        self.groups = sorted(groups)

    def _process_positionals(self, params: List[BasePositional]):
        var_nargs_param = None
        for param in params:
            if self.sub_command:
                raise CommandDefinitionError(
                    f'Positional param={param!r} may not follow the sub command {self.sub_command} - re-order the'
                    ' positionals, move it into the sub command(s), or convert it to an optional parameter'
                )
            elif var_nargs_param:
                raise CommandDefinitionError(
                    f'Additional Positional parameters cannot follow {var_nargs_param} because it accepts'
                    f' a variable number of arguments with no specific choices defined - param={param!r} is invalid'
                )

            if isinstance(param, (SubCommand, Action)):
                if self.action:  # self.sub_command being already defined is handled above
                    raise CommandDefinitionError(
                        f'Only 1 Action xor SubCommand is allowed in a given Command - {self.command.__name__} cannot'
                        f' contain both {self.action} and {param}'
                    )
                elif isinstance(param, SubCommand):
                    self.sub_command = param
                else:
                    self.action = param
                    if not param.has_choices:
                        raise CommandDefinitionError(f'No choices were registered for {self.action}')

            if param.nargs.variable and not param.has_choices:
                var_nargs_param = param

        self.positionals = params

    def _process_options(self, params: Collection[BaseOption]):
        parent = self.parent
        if parent:
            option_map = parent.option_map.copy()
            combo_option_map = parent.combo_option_map.copy()
            options = parent.options.copy()
        else:
            option_map = {}
            combo_option_map = {}
            options = []

        for param in params:
            options.append(param)
            opts = param.option_strs
            self._process_option_strs(param, 'long', opts.long, option_map, combo_option_map)
            self._process_option_strs(param, 'short', opts.short, option_map, combo_option_map)

        self.options = options
        self.option_map = option_map
        self._process_action_flags(options)
        self.combo_option_map = dict(sorted(combo_option_map.items(), key=lambda kv: (-len(kv[0]), kv[0])))  # noqa

        if self.config and self.config.ambiguous_short_combos == AmbiguousComboMode.STRICT:
            self._strict_ambiguous_short_combo_check()

    def _strict_ambiguous_short_combo_check(self):
        potentially_ambiguous_combo_options = self._potentially_ambiguous_combo_options
        if not potentially_ambiguous_combo_options:
            return

        param_conflicts_map = {
            param: singles.values() for param, singles in potentially_ambiguous_combo_options.values()
        }
        raise AmbiguousShortForm(param_conflicts_map)

    def _process_option_strs(
        self, param: BaseOption, opt_type: str, opt_strs: List[str], option_map: OptionMap, combo_option_map: OptionMap
    ):
        for opt in opt_strs:
            try:
                existing = option_map[opt]
            except KeyError:
                option_map[opt] = param
                if opt_type == 'short':
                    combo_option_map[opt[1:]] = param
            else:
                raise CommandDefinitionError(
                    f'{opt_type} option={opt!r} conflict for command={self.command!r} between {existing} and {param}'
                )

    def _process_action_flags(self, options: List[BaseOption]):
        action_flags = sorted((p for p in options if isinstance(p, ActionFlag)))
        grouped_ordered_flags = {True: defaultdict(list), False: defaultdict(list)}
        for param in action_flags:
            if param.func is None:
                raise ParameterDefinitionError(f'No function was registered for param={param!r}')
            grouped_ordered_flags[param.before_main][param.order].append(param)

        found_non_always = False
        invalid = {}
        for before_main, prio_params in grouped_ordered_flags.items():
            for prio, params in prio_params.items():
                param: ActionFlag = params[0]  # Don't pop and check `if params` - all are needed for the group check
                if found_non_always and param.always_available:
                    invalid[(before_main, prio)] = param
                elif not param.always_available:
                    found_non_always = True

                if len(params) > 1:
                    group = next((p.group for p in params if p.group), None)
                    if group and group.mutually_exclusive:
                        if any(p.group != group for p in params):
                            invalid[(before_main, prio)] = params
                    else:
                        invalid[(before_main, prio)] = params

        if invalid:
            raise CommandDefinitionError(
                f'ActionFlag parameters with the same before/after main setting must either have different order values'
                f' or be in a mutually exclusive ParamGroup, and always_available ActionFlags must all be ordered'
                f' before any that are not always available - invalid parameters: {invalid}'
            )

        n_before = len(grouped_ordered_flags[True])
        self.action_flags = action_flags
        self.split_action_flags = action_flags[:n_before], action_flags[n_before:]

    # endregion

    # region Option Processing

    def get_option_param_value_pairs(self, option: str) -> Optional[Tuple[BaseOption, ...]]:
        if option.startswith('---'):
            return None
        elif option.startswith('--'):
            try:
                opt, param, value = self.long_option_to_param_value_pair(option)
            except KeyError:
                return None
            else:
                return (param,)  # noqa
        elif option.startswith('-'):
            try:
                param_value_pairs = self.short_option_to_param_value_pairs(option)
            except KeyError:
                return None
            else:
                return tuple(param for opt, param, value in param_value_pairs)
        else:
            return None

    def long_option_to_param_value_pair(self, option: str) -> Tuple[str, BaseOption, Optional[str]]:
        try:
            return option, self.option_map[option], None
        except KeyError:
            if '=' in option:
                option, value = option.split('=', 1)
                return option, self.option_map[option], value
            else:
                raise

    def short_option_to_param_value_pairs(self, option: str) -> List[Tuple[str, BaseOption, Optional[str]]]:
        try:
            option, value = option.split('=', 1)
        except ValueError:
            value = None

        if len(option) > 2:  # 2 due to '-' prefix
            self._ensure_combo_is_unambiguous(option)

        try:
            param = self.option_map[option]
        except KeyError:
            if value is not None:
                raise
        else:
            return [(option, param, value)]

        key, value = option[1], option[2:]
        # value will never be empty if key is a valid option because by this point, option is not a short option
        combo_option_map = self.combo_option_map
        param = combo_option_map[key]
        if param.would_accept(value, short_combo=True):
            return [(key, param, value)]
        else:
            return [(c, combo_option_map[c], None) for c in option[1:]]

    def _ensure_combo_is_unambiguous(self, option: str):
        # Called by short_option_to_param_value_pairs after ensuring the length is > 1
        acm = AmbiguousComboMode.PERMISSIVE if not self.config else self.config.ambiguous_short_combos
        if acm == AmbiguousComboMode.IGNORE:
            return

        to_check = option[1:]  # Strip leading '-'
        potentially_ambiguous_combo_options = self._potentially_ambiguous_combo_options
        if acm == AmbiguousComboMode.PERMISSIVE and to_check in potentially_ambiguous_combo_options:
            return  # Permissive mode allows exact matches of multi-char short forms

        ambiguous = set()
        for multi, (param, singles) in potentially_ambiguous_combo_options.items():
            if multi in to_check:
                ambiguous.add(param)
                ambiguous.update(p for c, p in singles.items() if c in to_check)

        if ambiguous:
            raise AmbiguousCombo(ambiguous, option)

    def find_option_that_accepts_values(self, option: str) -> Optional[BaseOption]:
        if option.startswith('--'):
            param = self.long_option_to_param_value_pair(option)[1]
            if param.accepts_values:
                return param
        elif option.startswith('-'):
            for _, param, _ in self.short_option_to_param_value_pairs(option):
                if param.accepts_values:
                    return param
        else:
            raise ValueError(f'Invalid option={option!r}')
        return None

    def find_nested_option_that_accepts_values(self, option: str) -> Optional[BaseOption]:
        if not self.sub_command:
            return None

        for choice in self.sub_command.choices.values():
            command = choice.target
            params = command.__class__.params(command)
            try:
                param = params.find_option_that_accepts_values(option)
            except KeyError:
                pass
            else:
                if param is not None:
                    return param

            param = params.find_nested_option_that_accepts_values(option)
            if param is not None:
                return param

        return None

    def find_nested_pass_thru(self) -> Optional[PassThru]:
        if not self.sub_command:
            return None

        for choice in self.sub_command.choices.values():
            command = choice.target
            params = command.__class__.params(command)
            if params._pass_thru:
                return params._pass_thru

        return None

    # endregion

    def validate_groups(self):
        exc = None
        for group in self.groups:
            try:
                group.validate()
            except ParamsMissing as e:  # Let ParamConflict propagate before ParamsMissing
                if exc is None:
                    exc = e

        if exc is not None:
            raise exc

    def try_env_params(self, ctx: Context) -> Iterator[Option]:
        for param in self.options:
            try:
                param.env_var  # noqa
            except AttributeError:
                pass
            else:
                if ctx.num_provided(param) == 0:
                    yield param

    def required_check_params(self) -> Iterator[Parameter]:
        ignore = SubCommand
        yield from (p for p in self.positionals if p.required and not p.group and not isinstance(p, ignore))
        yield from (p for p in self.options if p.required and not p.group)
        pass_thru = self._pass_thru
        if pass_thru and pass_thru.required and not pass_thru.group:
            yield pass_thru


def _find_groups(groups: Set[ParamGroup], param: ParamBase):
    group = param.group
    while group:
        groups.add(group)
        group = group.group
